import numpy as np
import matplotlib.pyplot as plt


__all__ = ('error_plot',)


def error_plot(network, logx=False, ax=None, show=True):
    """
    Makes line plot that shows training progress. x-axis
    is an epoch number and y-axis is an error.

    Parameters
    ----------
    logx : bool
        Parameter set up logarithmic scale to x-axis.
        Defaults to ``False``.

    ax : object or None
        Matplotlib axis object. ``None`` values means that axis equal
        to the current one (the same as ``ax = plt.gca()``).
        Defaults to ``None``.

    show : bool
        If parameter is equal to ``True`` then plot will be
        displayed. Defaults to ``True``.

    Returns
    -------
    object
        Matplotlib axis instance.

    Examples
    --------
    >>> from neupy import algorithms, plots
    >>>
    >>> gdnet = algorithms.GradientDescent((2, 3, 1))
    >>> gdnet.train(x_train, y_train, x_test, y_test, epochs=100)
    >>>
    >>> plots.error_plot(gdnet)
    """
    if ax is None:
        ax = plt.gca()

    if not network.errors:
        network.logs.warning("There is no data to plot")
        return ax

    train_errors = network.errors.normalized()
    validation_errors = network.validation_errors.normalized()

    if len(train_errors) != len(validation_errors):
        network.logs.warning("Mismatch in number of training and validation "
                             "errors. Validation error will be ignored.")
        validation_errors = []

    if all(err is None for err in validation_errors):
        validation_errors = []

    errors_range = np.arange(len(train_errors))
    plot_function = ax.semilogx if logx else ax.plot

    line_error_in, = plot_function(errors_range, train_errors)

    if validation_errors:
        line_error_out, = plot_function(errors_range, validation_errors)
        ax.legend(
            [line_error_in, line_error_out],
            ['Train', 'Validation']
        )

    ax.set_title('Training perfomance')
    ax.set_ylim(bottom=0)

    ax.set_ylabel('Error')
    ax.set_xlabel('Epoch')

    if show:
        plt.show()

    return ax



import theano
import theano.tensor as T
import numpy as np

from neupy.utils import asfloat
from neupy.core.properties import ProperFractionProperty, NumberProperty
from .base import MinibatchGradientDescent


__all__ = ('Adam',)


class Adam(MinibatchGradientDescent):
    """
    Adam algorithm.

    Parameters
    ----------
    beta1 : float
        Decay rate. Value need to be between ``0`` and ``1``.
        Defaults to ``0.95``.

    beta2 : float
        Decay rate. Value need to be between ``0`` and ``1``.
        Defaults to ``0.95``.

    epsilon : float
        Value need to be greater than ``0``. Defaults to ``1e-5``.

    step : float
        Learning rate, defaults to ``0.001``.

    {MinibatchGradientDescent.batch_size}

    {GradientDescent.addons}

    {ConstructibleNetwork.connection}

    {ConstructibleNetwork.error}

    {BaseNetwork.show_epoch}

    {BaseNetwork.shuffle_data}

    {BaseNetwork.epoch_end_signal}

    {BaseNetwork.train_end_signal}

    {Verbose.verbose}

    Attributes
    ----------
    {MinibatchGradientDescent.Attributes}

    Methods
    -------
    {MinibatchGradientDescent.Methods}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> mnet = algorithms.Adam((2, 3, 1))
    >>> mnet.train(x_train, y_train)
    """
    step = NumberProperty(default=0.001, minval=0)
    beta1 = ProperFractionProperty(default=0.9)
    beta2 = ProperFractionProperty(default=0.999)
    epsilon = NumberProperty(default=1e-7, minval=0)

    def init_param_updates(self, layer, parameter):
        epoch = self.variables.epoch

        parameter_shape = T.shape(parameter).eval()
        prev_first_moment = theano.shared(
            name="{}/prev-first-moment".format(parameter.name),
            value=asfloat(np.zeros(parameter_shape)),
        )
        prev_second_moment = theano.shared(
            name="{}/prev-second-moment".format(parameter.name),
            value=asfloat(np.zeros(parameter_shape)),
        )

        step = self.variables.step
        gradient = T.grad(self.variables.error_func, wrt=parameter)

        first_moment = (
            self.beta1 * prev_first_moment +
            (1. - self.beta1) * gradient)
        second_moment = (
            self.beta2 * prev_second_moment +
            (1. - self.beta2) * gradient ** 2
        )

        first_moment_bias_corrected = first_moment / (1. - self.beta1 ** epoch)
        second_moment_bias_corrected = (
            second_moment / (1. - self.beta2 ** epoch))

        parameter_delta = first_moment_bias_corrected * (
            T.sqrt(second_moment_bias_corrected) + self.epsilon
        )

        return [
            (prev_first_moment, first_moment),
            (prev_second_moment, second_moment),
            (parameter, parameter - step * parameter_delta),
        ]



import theano.tensor as T

from neupy.core.properties import NumberProperty
from .base import WeightUpdateConfigurable


__all__ = ('MaxNormRegularization',)


def max_norm_clip(array, max_norm):
    """
    Clips norm of the array in case if it will exceed
    specified limit.

    Parameters
    ----------
    array : array-like
        Any array-like object.

    max_norm : float
        Maximum possible norm value for the input array. Array will
        be clipped in case if its norm will be greater than
        specified limit.
    """
    array_norm = T.sqrt(T.sum(array ** 2))
    return T.switch(
        T.ge(array_norm, max_norm),
        T.mul(max_norm, array) / array_norm,
        array)


class MaxNormRegularization(WeightUpdateConfigurable):
    """
    Max-norm regularization algorithm will clip norm of the
    parameter in case if it will exceed maximum limit.

    .. code-block:: python

        if norm(weight) > max_norm:
            weight = max_norm * weight / norm(weight)

    .. raw:: html

        <br>

    Warns
    -----
    {WeightUpdateConfigurable.Warns}

    Parameters
    ----------
    max_norm : int, float
        Any parameter that has norm greater than this value
        will be clipped. Defaults to ``10``.

    Examples
    --------
    >>> from neupy import algorithms
    >>> bpnet = algorithms.GradientDescent(
    ...     (2, 4, 1),
    ...     step=0.1,
    ...     max_norm=4,
    ...     addons=[algorithms.MaxNormRegularization]
    ... )

    References
    ----------
    [1] N. Srivastava, G. Hinton, A. Krizhevsky, I. Sutskever,
      R. Salakhutdinov. Dropout: A Simple Way to Prevent
      Neural Networks from Overfitting.
      http://jmlr.org/papers/volume15/srivastava14a/srivastava14a.pdf
    """
    max_norm = NumberProperty(default=10, minval=0)

    def init_param_updates(self, layer, parameter):
        updates = super(MaxNormRegularization, self).init_param_updates(
            layer, parameter)

        updates_mapper = dict(updates)
        updated_value = updates_mapper[parameter]
        updates_mapper[parameter] = max_norm_clip(updated_value, self.max_norm)

        return list(updates_mapper.items())



import theano
from theano.ifelse import ifelse
import numpy as np

from neupy.core.properties import (BoundedProperty,
                                   ProperFractionProperty)
from .base import SingleStepConfigurable


__all__ = ('ErrDiffStepUpdate',)


class ErrDiffStepUpdate(SingleStepConfigurable):
    """
    This algorithm make step update base on error difference between
    epochs.

    Parameters
    ----------
    update_for_smaller_error : float
        Multiplies this option to ``step`` in if the error
        was less than in previous epochs. Defaults to ``1.05``.
        Value can't be less than ``1``.

    update_for_bigger_error : float
        Multiplies this option to ``step`` in if the error
        was more than in previous epochs. Defaults to ``0.7``.

    error_difference : float
        The value indicates how many had to increase the
        error from the previous epochs that would produce
        reduction step. Defaults to ``1.04``.
        Value can't be less than ``1``.

    Warns
    -----
    {SingleStepConfigurable.Warns}

    Examples
    --------
    >>> from neupy import algorithms
    >>>
    >>> bpnet = algorithms.GradientDescent(
    ...     (2, 4, 1),
    ...     step=0.1,
    ...     verbose=False,
    ...     addons=[algorithms.ErrDiffStepUpdate]
    ... )
    """
    update_for_smaller_error = BoundedProperty(default=1.05, minval=1)
    update_for_bigger_error = ProperFractionProperty(default=0.7)
    error_difference = BoundedProperty(default=1.04, minval=1)

    def init_variables(self):
        self.variables.update(
            last_error=theano.shared(
                name='err-diff-step-update/last-error',
                value=np.nan
            ),
            previous_error=theano.shared(
                name='err-diff-step-update/previous-error',
                value=np.nan
            ),
        )
        super(ErrDiffStepUpdate, self).init_variables()

    def init_train_updates(self):
        updates = super(ErrDiffStepUpdate, self).init_train_updates()

        step = self.variables.step
        last_error = self.variables.last_error
        previous_error = self.variables.previous_error

        step_update_condition = ifelse(
            last_error < previous_error,
            self.update_for_smaller_error * step,
            ifelse(
                last_error > self.update_for_bigger_error * previous_error,
                self.update_for_bigger_error * step,
                step
            )

        )
        updates.append((step, step_update_condition))
        return updates

    def on_epoch_start_update(self, epoch):
        super(ErrDiffStepUpdate, self).on_epoch_start_update(epoch)

        previous_error = self.errors.previous()
        if previous_error:
            last_error = self.errors.last()
            self.variables.last_error.set_value(last_error)
            self.variables.previous_error.set_value(previous_error)



from neupy.algorithms.linear.base import BaseLinearNetwork


__all__ = ('LMS',)


class LMS(BaseLinearNetwork):
    """
    LMS Neural Network. Algorithm has several names,
    including the Widrow-Hoff or Delta rule.

    Parameters
    ----------
    {BaseLinearNetwork.connection}

    {BaseNetwork.step}

    {BaseNetwork.show_epoch}

    {BaseNetwork.shuffle_data}

    {BaseNetwork.epoch_end_signal}

    {BaseNetwork.train_end_signal}

    {Verbose.verbose}

    Methods
    -------
    {BaseSkeleton.predict}

    {ConstructibleNetwork.train}

    {BaseSkeleton.fit}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>>
    >>> input_data = np.array([[1, 0], [2, 2], [3, 3], [0, 0]])
    >>> target_data = np.array([[1], [0], [0], [1]])
    >>>
    >>> lmsnet = algorithms.LMS((2, 1), step=0.1)
    >>>
    >>> lmsnet.train(input_data, target_data, epochs=200)
    >>> lmsnet.predict(np.array([[4, 4], [0, 0]]))
    array([[0],
           [1]])

    See Also
    --------
    :network:`Perceptron` : Perceptron Neural Network.
    """
    def init_train_updates(self):
        layer = self.connection.output_layers[0]

        network_output = self.variables.network_output
        network_input = self.variables.network_input
        step = self.variables.step

        summated_output = network_input.dot(layer.weight) + layer.bias
        linear_error = summated_output - network_output

        normalized_input = network_input / network_input.norm(L=2)
        weight_delta = normalized_input.T.dot(linear_error)
        bias_delta = linear_error.sum(axis=0)

        return [
            (layer.weight, layer.weight - step * weight_delta),
            (layer.bias, layer.bias - step * bias_delta),
        ]



__all__ = ('SINGLE_STEP_UPDATE', 'MULTIPLE_STEP_UPDATE', 'WEIGHT_PENALTY',
           'addon_types', 'StepSelectionBuiltIn', 'NoMultipleStepSelection')


# Available add-on types
SINGLE_STEP_UPDATE = 1
MULTIPLE_STEP_UPDATE = 2
WEIGHT_PENALTY = 3


addon_types = {
    SINGLE_STEP_UPDATE: "Step update",
    WEIGHT_PENALTY: "Regularization",
}


class StepSelectionBuiltIn(object):
    """
    Mixin excludes add-ons that modify learning rate.
    """
    supported_addon_types = [WEIGHT_PENALTY]


class NoMultipleStepSelection(object):
    """
    Mixin excludes add-ons that use multiple learning rates for
    one neural network.
    """
    supported_addon_types = [SINGLE_STEP_UPDATE, WEIGHT_PENALTY]



import numpy as np
import theano.tensor as T

from neupy.utils import as_tuple
from neupy.core.properties import TypedListProperty
from .base import BaseLayer


__all__ = ('Reshape',)


class Reshape(BaseLayer):
    """
    Gives a new shape to an input value without changing
    its data.

    Parameters
    ----------
    shape : tuple or list
        New feature shape. ``None`` value means that feature
        will be flatten in 1D vector. If you need to get the
        output feature with more that 2 dimensions then you can
        set up new feature shape using tuples. Defaults to ``None``.

    {BaseLayer.Parameters}

    Methods
    -------
    {BaseLayer.Methods}

    Attributes
    ----------
    {BaseLayer.Attributes}

    Examples
    --------

    Covert 4D input to 2D

    >>> from neupy import layers
    >>>
    >>> connection = layers.join(
    ...     layers.Input((2, 5, 5)),
    ...     layers.Reshape()
    ... )
    >>>
    >>> print("Input shape: {{}}".format(connection.input_shape))
    Input shape: (2, 5, 5)
    >>>
    >>> print("Output shape: {{}}".format(connection.output_shape))
    Output shape: (50,)

    Convert 3D to 4D

    >>> from neupy import layers
    >>>
    >>> connection = layers.join(
    ...     layers.Input((5, 4)),
    ...     layers.Reshape((5, 2, 2))
    ... )
    >>>
    >>> print("Input shape: {{}}".format(connection.input_shape))
    Input shape: (5, 4)
    >>>
    >>> print("Output shape: {{}}".format(connection.output_shape))
    Output shape: (5, 2, 2)
    """
    shape = TypedListProperty()

    def __init__(self, shape=None, **options):
        if shape is not None:
            options['shape'] = shape
        super(Reshape, self).__init__(**options)

    @property
    def output_shape(self):
        if self.shape is not None:
            return as_tuple(self.shape)

        n_output_features = np.prod(self.input_shape)
        return as_tuple(n_output_features)

    def output(self, input_value):
        """
        Reshape the feature space for the input value.

        Parameters
        ----------
        input_value : array-like or Theano variable
        """
        n_samples = input_value.shape[0]
        output_shape = as_tuple(n_samples, self.output_shape)
        return T.reshape(input_value, output_shape)



import numpy as np

from neupy.utils import format_data
from neupy.core.properties import IntProperty
from neupy.algorithms.base import BaseNetwork


__all__ = ('CMAC',)


class CMAC(BaseNetwork):
    """
    CMAC Network based on memory.

    Notes
    -----
    - Network always use Mean Absolute Error (MAE).
    - Network works for multi dimensional target values.

    Parameters
    ----------
    quantization : int
        Network transforms every input to discrete value.
        Quantization value contol number of total possible
        categories after quantization, defaults to ``10``.

    associative_unit_size : int
        Number of associative blocks in memory, defaults to ``2``.

    {BaseNetwork.Parameters}

    Attributes
    ----------
    weight : dict
        Network's weight that contains memorized patterns.

    Methods
    -------
    {BaseSkeleton.predict}

    train(input_train, target_train, input_test=None, target_test=None,\
    epochs=100, epsilon=None)
        Train network. You can control network's training procedure
        with ``epochs`` and ``epsilon`` parameters.
        The ``input_test`` and ``target_test`` should be presented
        both in case of you need to validate network's training
        after each iteration.

    {BaseSkeleton.fit}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy.algorithms import CMAC
    >>>
    >>> train_space = np.linspace(0, 2 * np.pi, 100)
    >>> test_space = np.linspace(np.pi, 2 * np.pi, 50)
    >>>
    >>> input_train = np.reshape(train_space, (100, 1))
    >>> input_test = np.reshape(test_space, (50, 1))
    >>>
    >>> target_train = np.sin(input_train)
    >>> target_test = np.sin(input_test)
    >>>
    >>> cmac = CMAC(
    ...     quantization=100,
    ...     associative_unit_size=32,
    ...     step=0.2,
    ... )
    ...
    >>> cmac.train(input_train, target_train, epochs=100)
    >>>
    >>> predicted_test = cmac.predict(input_test)
    >>> cmac.error(target_test, predicted_test)
    0.0023639417543036569
    """
    quantization = IntProperty(default=10, minval=1)
    associative_unit_size = IntProperty(default=2, minval=2)

    def __init__(self, **options):
        self.weight = {}
        super(CMAC, self).__init__(**options)

    def predict(self, input_data):
        input_data = format_data(input_data)

        get_memory_coords = self.get_memory_coords
        get_result_by_coords = self.get_result_by_coords
        predicted = []

        for input_sample in self.quantize(input_data):
            coords = get_memory_coords(input_sample)
            predicted.append(get_result_by_coords(coords))

        return np.array(predicted)

    def get_result_by_coords(self, coords):
        return sum(
            self.weight.setdefault(coord, 0) for coord in coords
        ) / self.associative_unit_size

    def get_memory_coords(self, quantized_value):
        assoc_unit_size = self.associative_unit_size

        for i in range(assoc_unit_size):
            point = ((quantized_value + i) / assoc_unit_size).astype(int)
            yield tuple(np.concatenate([point, [i]]))

    def quantize(self, input_data):
        return (input_data * self.quantization).astype(int)

    def train_epoch(self, input_train, target_train):
        get_memory_coords = self.get_memory_coords
        get_result_by_coords = self.get_result_by_coords
        weight = self.weight
        step = self.step

        n_samples = input_train.shape[0]
        quantized_input = self.quantize(input_train)
        errors = 0

        for input_sample, target_sample in zip(quantized_input, target_train):
            coords = list(get_memory_coords(input_sample))
            predicted = get_result_by_coords(coords)

            error = target_sample - predicted
            for coord in coords:
                weight[coord] += step * error

            errors += abs(error)

        return errors / n_samples

    def prediction_error(self, input_data, target_data):
        predicted = self.predict(input_data)
        return np.mean(np.abs(predicted - target_data))

    def train(self, input_train, target_train, input_test=None,
              target_test=None, epochs=100, epsilon=None,
              summary='table'):

        is_test_data_partialy_missed = (
            (input_test is None and target_test is not None) or
            (input_test is not None and target_test is None)
        )

        if is_test_data_partialy_missed:
            raise ValueError("Input and target test samples are missed. "
                             "They must be defined together or none of them.")

        input_train = format_data(input_train)
        target_train = format_data(target_train)

        if input_test is not None:
            input_test = format_data(input_test)

        if target_test is not None:
            target_test = format_data(target_test)

        return super(CMAC, self).train(
            input_train=input_train, target_train=target_train,
            input_test=input_test, target_test=target_test,
            epochs=epochs, epsilon=epsilon,
            summary=summary
        )



from neupy.core.base import BaseSkeleton
from neupy.core.config import Configurable


__all__ = ('BaseEnsemble',)


class BaseEnsemble(BaseSkeleton, Configurable):
    """
    Base class for ensemble algorithms.

    Parameters
    ----------
    networks : list or tuple
        List of networks.

    Raises
    ------
    ValueError
    """
    def __init__(self, networks):
        self.networks = networks
        n_networks = len(self.networks)

        if n_networks < 2:
            raise ValueError("Ensemble algorithm should have at least "
                             "2 networks, got {}.".format(n_networks))

    def __repr__(self):
        return "{classname}(networks=[\n    {networks}\n])".format(
            classname=self.__class__.__name__,
            networks=',\n    '.join(map(repr, self.networks))
        )



"""
Main source code from Pylearn2 library:
https://github.com/lisa-lab/pylearn2/blob/master/pylearn2/\
optimization/linesearch.py
"""
import theano
import theano.tensor as T
from theano.ifelse import ifelse

from neupy.utils import asfloat


one = T.constant(asfloat(1))
zero = T.constant(asfloat(0))

theano_true = T.constant(1)
theano_false = T.constant(0)


def sequential_or(*conditions):
    """
    Use ``or`` operator between all conditions. Function is just
    a syntax sugar that make long Theano logical conditions looks
    less ugly.

    Parameters
    ----------
    *conditions
        Conditions that returns ``True`` or ``False``
    """
    first_condition, other_conditions = conditions[0], conditions[1:]
    if not other_conditions:
        return first_condition
    return T.or_(first_condition, sequential_or(*other_conditions))


def sequential_and(*conditions):
    """
    Use ``and`` operator between all conditions. Function is just
    a syntax sugar that make long Theano logical conditions looks
    less ugly.

    Parameters
    ----------
    *conditions
        Conditions that returns ``True`` or ``False``
    """
    first_condition, other_conditions = conditions[0], conditions[1:]
    if not other_conditions:
        return first_condition
    return T.and_(first_condition, sequential_and(*other_conditions))


def bitwise_not(condition):
    """
    Bitwise not compatible with Theano version 0.8.* and 0.9.*

    Parameters
    ----------
    condition : bool, int
    """
    return T.neq(condition, 1)


def line_search(f, f_deriv, maxiter=20, c1=1e-4, c2=0.9):
    """
    Find ``x`` that satisfies strong Wolfe conditions.
    ``x > 0`` is assumed to be a descent direction.

    Parameters
    ----------
    f : callable f(x)
        Objective scalar function.
    f_deriv : callable f'(x)
        Objective function derivative (can be None)
    maxiter : int
        Maximum number of iterations.
    c1 : float
        Parameter for Armijo condition rule.
    c2 : float
        Parameter for curvature condition rule.

    Returns
    -------
    Theano object
        Value ``x`` that satisfies strong Wolfe conditions and
        minimize function ``f``.

    Notes
    -----
    Uses the line search algorithm to enforce strong Wolfe
    conditions.  See Wright and Nocedal, 'Numerical Optimization',
    1999, pg. 59-60.
    For the zoom phase it uses an algorithm by [...].
    """

    if not 0 < c1 < 1:
        raise ValueError("c1 should be a float between 0 and 1")

    if not 0 < c2 < 1:
        raise ValueError("c2 should be a float between 0 and 1")

    if c2 < c1:
        raise ValueError("c2 needs to be greater than c1")

    if maxiter <= 0:
        raise ValueError("maxiter needs to be greater than 0")

    c1, c2 = asfloat(c1), asfloat(c2)

    def search_iteration_step(x_previous, x_current, y_previous, y_current,
                              y_deriv_previous, is_first_iteration, x_star):

        y_deriv_current = f_deriv(x_current)

        x_new = x_current * asfloat(2)
        y_new = f(x_new)

        condition1 = T.or_(
            y_current > (y0 + c1 * x_current * y_deriv_0),
            T.and_(
                y_current >= y_previous,
                bitwise_not(is_first_iteration),
            )
        )
        condition2 = T.abs_(y_deriv_current) <= -c2 * y_deriv_0
        condition3 = y_deriv_current >= zero

        x_star = ifelse(
            condition1,
            zoom(
                x_previous, x_current, y_previous,
                y_current, y_deriv_previous,
                f, f_deriv, y0, y_deriv_0, c1, c2
            ),
            ifelse(
                condition2,
                x_current,
                ifelse(
                    condition3,
                    zoom(
                        x_current, x_previous, y_current,
                        y_previous, y_deriv_current,
                        f, f_deriv, y0, y_deriv_0, c1, c2
                    ),
                    x_new,
                ),
            ),
        )
        y_deriv_previous_new = ifelse(
            condition1,
            y_deriv_previous,
            y_deriv_current
        )

        is_any_condition_satisfied = sequential_or(condition1, condition2,
                                                   condition3)
        y_current_new = ifelse(
            is_any_condition_satisfied,
            y_current,
            y_new
        )
        return (
            [
                x_current, x_new, y_current, y_current_new,
                y_deriv_previous_new, theano_false, x_star
            ],
            theano.scan_module.scan_utils.until(
                sequential_or(
                    T.eq(x_new, zero),
                    is_any_condition_satisfied,
                )
            )
        )

    x0, x1 = zero, one
    y0, y1 = f(x0), f(x1)
    y_deriv_0 = f_deriv(x0)

    c1 = T.as_tensor_variable(c1)
    c2 = T.as_tensor_variable(c2)

    outs, _ = theano.scan(
        search_iteration_step,
        outputs_info=[x0, x1, y0, y1, y_deriv_0, theano_true, zero],
        n_steps=maxiter
    )
    x_star = outs[-1][-1]

    return x_star


def quadratic_minimizer(x_a, y_a, y_prime_a, x_b, y_b, bound_size_ratio=0.1):
    """
    Finds the minimizer for a quadratic polynomial that
    goes through the points (x_a, y_a), (x_b, y_b) with derivative
    at x_a of y_prime_a.

    Parameters
    ----------
    x_a : float or theano variable
        Left point ``a`` in the ``x`` axis.
    y_a : float or theano variable
        Output from function ``y`` at point ``a``.
    y_prime_a : float or theano variable
        Output from function ``y'`` (``y`` derivative) at
        point ``a``.
    x_b : float or theano variable
        Right point ``a`` in the ``x`` axis.
    y_b : float or theano variable
        Output from function ``y`` at point ``b``.
    bound_size_ratio : float
        Value control acceptable bounds for interpolation. If value
        close to one of the points interpolation result will be ignored.
        The bigger ratio, the more likely to reject interpolation.
        Value needs to be between ``0`` and ``1``. Defaults to ``0.1``.

    Returns
    -------
    object
        Theano variable that after evaluation is equal to
        point ``x`` which is minimizer for quadratic function.
    """

    if not 0 <= bound_size_ratio < 1:
        raise ValueError("Value ``bound_size_ratio`` need to be a float "
                         "between 0 and 1, got {}".format(bound_size_ratio))

    # The main formula works for the region [0, a] we need to
    # shift function to the left side and put point ``a``
    # at ``0`` position.
    x_range = x_b - x_a
    coef = (y_b - y_a - y_prime_a * x_range) / (x_range ** asfloat(2))
    minimizer = -y_prime_a / (asfloat(2) * coef) + x_a
    bound_size_ratio = asfloat(bound_size_ratio)

    return T.switch(
        sequential_or(
            # Handle bad cases
            T.eq(x_range, zero),
            coef <= zero,

            T.isnan(minimizer),
            T.gt(minimizer, x_b - bound_size_ratio * x_range),
            T.lt(minimizer, x_a + bound_size_ratio * x_range),
        ),
        x_a + asfloat(0.5) * x_range,
        # Since we shifted funciton to the left, we need to shift
        # the result to the right to make it correct for
        # the specified region. That's why we are adding ``x_a``
        # at the end.
        -y_prime_a / (asfloat(2) * coef) + x_a
    )


def cubic_minimizer(x_a, y_a, y_prime_a, x_b, y_b, x_c, y_c,
                    bound_size_ratio=0.2):
    """
    Finds the minimizer for a cubic polynomial that goes
    through the points (x_a, y_a), (x_b, y_b), and (x_c, y_c)
    with derivative at ``x_a`` of y_prime_a.

    Parameters
    ----------
    x_a : float or theano variable
        First point ``a`` in the ``x`` axis.
    y_a : float or theano variable
        Output from function ``y`` at point ``a``.
    y_prime_a : float or theano variable
        Output from function ``y'`` (``y`` derivative) at
        point ``a``.
    x_b : float or theano variable
        Second point ``b`` in the ``x`` axis.
    y_b : float or theano variable
        Output from function ``y`` at point ``b``.
    x_c : float or theano variable
        Third point ``c`` in the ``x`` axis.
    y_c : float or theano variable
        Output from function ``y`` at point ``c``.
    bound_size_ratio : float
        Value control acceptable bounds for interpolation. If
        value is close to one of the points than interpolation
        result will be ignored. The bigger the ratio, the more
        likely it's going to reject interpolation. Value needs
        to be between ``0`` and ``1``. Defaults to ``0.1``.

    Returns
    -------
    object
        Theano variable that after evaluation is equal to
        the point ``x`` which is a minimizer for the cubic function.
    """

    if not 0 <= bound_size_ratio < 1:
        raise ValueError("The `bound_size_ratio` value should be a float "
                         "number between 0 and 1, got {}"
                         "".format(bound_size_ratio))

    bound_size_ratio = asfloat(bound_size_ratio)

    from_a2b_dist = x_b - x_a
    from_a2c_dist = x_c - x_a

    denominator = (
        (from_a2b_dist * from_a2c_dist) ** asfloat(2) *
        (from_a2b_dist - from_a2c_dist)
    )
    tau_ab = y_b - y_a - y_prime_a * from_a2b_dist
    tau_ac = y_c - y_a - y_prime_a * from_a2c_dist

    alpha = (
        from_a2c_dist ** asfloat(2) * tau_ab -
        from_a2b_dist ** asfloat(2) * tau_ac
    ) / denominator
    beta = (
        from_a2b_dist ** asfloat(3) * tau_ac -
        from_a2c_dist ** asfloat(3) * tau_ab
    ) / denominator
    radical = beta ** asfloat(2) - asfloat(3) * alpha * y_prime_a

    minimizer = x_a + (-beta + T.sqrt(radical)) / (asfloat(3) * alpha)

    return ifelse(
        sequential_or(
            # Handle bad cases
            radical < zero,

            T.eq(x_a, x_b),
            T.eq(x_a, x_c),
            T.eq(x_b, x_c),
            T.eq(alpha, zero),

            T.isnan(minimizer),
            T.gt(minimizer, x_b - bound_size_ratio * from_a2b_dist),
            T.lt(minimizer, x_a + bound_size_ratio * from_a2b_dist),
        ),
        quadratic_minimizer(x_a, y_a, y_prime_a, x_b, y_b),
        minimizer
    )


def zoom(x_low, x_high, y_low, y_high, y_deriv_low,
         f, f_deriv, y0, y_deriv_0, c1, c2, maxiter=10):
    """
    Notes
    -----
    Part of the optimization algorithm in `scalar_search_wolfe2`.

    Parameters
    ----------
    x_low : float
        Step size
    x_high : float
        Step size
    y_low : float
        Value of f at x_low
    y_high : float
        Value of f at x_high
    y_deriv_low : float
        Value of derivative at x_low
    f : callable f(x)
        Generates computational graph
    f_deriv : callable f'(x)
        Generates computational graph
    y0 : float
        Value of f for ``x = 0``
    y_deriv_0 : float
        Value of the derivative for ``x = 0``
    c1 : float
        Parameter for Armijo condition rule.
    c2 : float
        Parameter for curvature condition rule.
    """

    def zoom_itertion_step(x_low, y_low, y_deriv_low, x_high, y_high,
                           x_recent, y_recent, x_star):
        x_new = cubic_minimizer(x_low, y_low, y_deriv_low,
                                x_high, y_high,
                                x_recent, y_recent)

        y_new = f(x_new)
        y_deriv_new = f_deriv(x_new)

        stop_loop_rule = sequential_and(
            y_new <= y0 + c1 * x_new * y_deriv_0,
            y_new < y_low,
            abs(y_deriv_new) <= -c2 * y_deriv_0,
        )

        condition1 = T.or_(
            y_new > y0 + c1 * x_new * y_deriv_0,
            y_new >= y_low
        )
        condition2 = y_deriv_new * (x_high - x_low) >= zero

        y_recent, x_recent, x_high, y_high = ifelse(
            condition1,
            [y_high, x_high, x_new, y_new],
            ifelse(
                condition2,
                [y_high, x_high, x_low, y_low],
                [y_low, x_low, x_high, y_high],
            )
        )

        x_low, y_low, y_deriv_low = ifelse(
            condition1,
            [x_low, y_low, y_deriv_low],
            [x_new, y_new, y_deriv_new],
        )
        x_star = x_new

        return (
            [
                x_low, y_low, y_deriv_low,
                x_high, y_high,
                y_recent, x_recent,
                x_star
            ],
            theano.scan_module.scan_utils.until(stop_loop_rule)
        )

    x_recent = zero
    y_recent = y0

    outs, _ = theano.scan(
        zoom_itertion_step,
        outputs_info=[
            x_low, y_low, y_deriv_low,
            x_high, y_high,
            x_recent, y_recent,
            zero,
        ],
        n_steps=maxiter
    )

    return outs[-1][-1]



import math

import numpy as np
from numpy.core.umath_tests import inner1d


__all__ = ('pdf_between_data',)


def pdf_between_data(train_data, input_data, std):
    """
    Compute PDF between two samples.

    Parameters
    ----------
    train_data : array
        Training dataset.

    input_data : array
        Input dataset

    std : float
        Standard deviation for Probability Density
        Function (PDF).

    Returns
    -------
    array-like
    """
    n_train_samples = train_data.shape[0]
    n_samples = input_data.shape[0]

    results = np.zeros((n_train_samples, n_samples))
    variance = std ** 2
    const = std * math.sqrt(2 * math.pi)

    for i, input_row in enumerate(input_data):
        inputs = np.tile(input_row, (n_train_samples, 1))
        class_difference = (train_data - inputs)
        total_distance = inner1d(class_difference, class_difference)
        results[:, i] = np.exp(-total_distance / variance) / const

    return results



import theano
import theano.tensor as T
import numpy as np

from neupy.utils import asfloat
from neupy.core.properties import ProperFractionProperty, NumberProperty
from .base import MinibatchGradientDescent


__all__ = ('Adamax',)


class Adamax(MinibatchGradientDescent):
    """
    AdaMax algorithm.

    Parameters
    ----------
    beta1 : float
        Decay rate. Value need to be between ``0`` and ``1``.
        Defaults to ``0.95``.

    beta2 : float
        Decay rate. Value need to be between ``0`` and ``1``.
        Defaults to ``0.95``.

    epsilon : float
        Value need to be greater than ``0``. Defaults to ``1e-5``.

    step : float
        Learning rate, defaults to ``0.001``.

    {MinibatchGradientDescent.batch_size}

    {GradientDescent.addons}

    {ConstructibleNetwork.connection}

    {ConstructibleNetwork.error}

    {BaseNetwork.show_epoch}

    {BaseNetwork.shuffle_data}

    {BaseNetwork.epoch_end_signal}

    {BaseNetwork.train_end_signal}

    {Verbose.verbose}

    Attributes
    ----------
    {MinibatchGradientDescent.Attributes}

    Methods
    -------
    {MinibatchGradientDescent.Methods}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> mnet = algorithms.Adamax((2, 3, 1))
    >>> mnet.train(x_train, y_train)
    """
    step = NumberProperty(default=0.001, minval=0)
    beta1 = ProperFractionProperty(default=0.9)
    beta2 = ProperFractionProperty(default=0.999)
    epsilon = NumberProperty(default=1e-8, minval=0)

    def init_param_updates(self, layer, parameter):
        epoch = self.variables.epoch
        step = self.variables.step
        beta1 = self.beta1
        beta2 = self.beta2

        parameter_shape = T.shape(parameter).eval()
        prev_first_moment = theano.shared(
            name="{}/prev-first-moment".format(parameter.name),
            value=asfloat(np.zeros(parameter_shape)),
        )
        prev_weighted_inf_norm = theano.shared(
            name="{}/prev-weighted-inf-norm".format(parameter.name),
            value=asfloat(np.zeros(parameter_shape)),
        )

        gradient = T.grad(self.variables.error_func, wrt=parameter)

        first_moment = beta1 * prev_first_moment + (1 - beta1) * gradient
        weighted_inf_norm = T.maximum(beta2 * prev_weighted_inf_norm,
                                      T.abs_(gradient))

        parameter_delta = (
            (1 / (1 - beta1 ** epoch)) *
            (first_moment / (weighted_inf_norm + self.epsilon))
        )

        return [
            (prev_first_moment, first_moment),
            (prev_weighted_inf_norm, weighted_inf_norm),
            (parameter, parameter - step * parameter_delta),
        ]



import seaborn as sns
import pandas as pd
import matplotlib.pyplot as plt
from sklearn import datasets
from neupy import algorithms, environment


environment.reproducible()


def plot_scattermatrix(data, target):
    df = pd.DataFrame(data)
    df['target'] = target
    return sns.pairplot(df, hue='target')


if __name__ == '__main__':
    dataset = datasets.load_iris()
    data, target = dataset.data, dataset.target

    lvqnet = algorithms.LVQ3(
        # number of features
        n_inputs=4,

        # number of data points that we want
        # to have at the end
        n_subclasses=30,

        # number of classes
        n_classes=3,

        verbose=True,
        show_epoch=20,

        step=0.001,
        n_updates_to_stepdrop=150 * 100,
    )
    lvqnet.train(data, target, epochs=100)

    plot_scattermatrix(data, target)
    plot_scattermatrix(data=lvqnet.weight, target=lvqnet.subclass_to_class)
    plt.show()



import theano
import theano.tensor as T
from theano.ifelse import ifelse
from theano.tensor import slinalg
import numpy as np

from neupy.utils import asfloat
from neupy.core.properties import (BoundedProperty, ChoiceProperty,
                                   WithdrawProperty)
from neupy.algorithms import GradientDescent
from neupy.algorithms.gd import StepSelectionBuiltIn, errors
from neupy.algorithms.utils import parameter_values, setup_parameter_updates


__all__ = ('LevenbergMarquardt',)


def compute_jacobian(errors, parameters):
    """
    Compute jacobian.

    Parameters
    ----------
    errors : Theano variable
        Computed MSE for each sample separetly.

    parameters : list of Theano variable
        Neural network parameters (e.g. weights, biases).

    Returns
    -------
    Theano variable
    """
    n_samples = errors.shape[0]
    J = T.jacobian(errors, wrt=parameters)

    jacobians = []
    for jacobian, parameter in zip(J, parameters):
        jacobian = jacobian.reshape((n_samples, parameter.size))
        jacobians.append(jacobian)

    return T.concatenate(jacobians, axis=1)


class LevenbergMarquardt(StepSelectionBuiltIn, GradientDescent):
    """
    Levenberg-Marquardt algorithm.

    Notes
    -----
    - Network minimizes only Mean Squared Error function.
    - Efficient for small training datasets, because it
      computes gradient per each sample separately.
    - Efficient for small-sized networks.

    Parameters
    ----------
    {GradientDescent.connection}

    mu : float
        Control invertion for J.T * J matrix, defaults to `0.1`.

    mu_update_factor : float
        Factor to decrease the mu if update decrese the error, otherwise
        increse mu by the same factor. Defaults to ``1.2``

    error : {{``mse``}}
        Levenberg-Marquardt works only for quadratic functions.
        Defaults to ``mse``.

    {GradientDescent.show_epoch}

    {GradientDescent.shuffle_data}

    {GradientDescent.epoch_end_signal}

    {GradientDescent.train_end_signal}

    {GradientDescent.verbose}

    {GradientDescent.addons}

    Attributes
    ----------
    {GradientDescent.Attributes}

    Methods
    -------
    {GradientDescent.Methods}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> lmnet = algorithms.LevenbergMarquardt((2, 3, 1))
    >>> lmnet.train(x_train, y_train)

    See Also
    --------
    :network:`GradientDescent` : GradientDescent algorithm.
    """
    mu = BoundedProperty(default=0.01, minval=0)
    mu_update_factor = BoundedProperty(default=1.2, minval=1)
    error = ChoiceProperty(default='mse', choices={'mse': errors.mse})

    step = WithdrawProperty()

    def init_variables(self):
        super(LevenbergMarquardt, self).init_variables()
        self.variables.update(
            mu=theano.shared(name='lev-marq/mu', value=asfloat(self.mu)),
            last_error=theano.shared(name='lev-marq/last-error', value=np.nan),
        )

    def init_train_updates(self):
        network_output = self.variables.network_output
        prediction_func = self.variables.train_prediction_func
        last_error = self.variables.last_error
        error_func = self.variables.error_func
        mu = self.variables.mu

        new_mu = ifelse(
            T.lt(last_error, error_func),
            mu * self.mu_update_factor,
            mu / self.mu_update_factor,
        )

        se_for_each_sample = (
            (network_output - prediction_func) ** 2
        ).ravel()

        params = parameter_values(self.connection)
        param_vector = T.concatenate([param.flatten() for param in params])

        J = compute_jacobian(se_for_each_sample, params)
        n_params = J.shape[1]

        updated_params = param_vector - slinalg.solve(
            J.T.dot(J) + new_mu * T.eye(n_params),
            J.T.dot(se_for_each_sample)
        )

        updates = [(mu, new_mu)]
        parameter_updates = setup_parameter_updates(params, updated_params)
        updates.extend(parameter_updates)

        return updates

    def on_epoch_start_update(self, epoch):
        super(LevenbergMarquardt, self).on_epoch_start_update(epoch)

        last_error = self.errors.last()
        if last_error is not None:
            self.variables.last_error.set_value(last_error)



from numpy import dot

from neupy.utils import format_data
from neupy.exceptions import NotTrained
from neupy.core.properties import BoundedProperty
from neupy.algorithms.base import BaseNetwork
from .learning import LazyLearningMixin
from .utils import pdf_between_data


__all__ = ('GRNN',)


class GRNN(LazyLearningMixin, BaseNetwork):
    """
    Generalized Regression Neural Network (GRNN). Network applies
    only to the regression problems.

    Parameters
    ----------
    std : float
        Standard deviation for PDF function, defaults to ``0.1``.
        If your input features have high values than standard
        deviation should also be high. For instance, if input features
        from range ``[0, 20]`` that standard deviation should be
        also a big value like ``10`` or ``15``. Small values will
        lead to bad prediction.

    {Verbose.verbose}

    Notes
    -----
    - GRNN Network is sensitive for cases when one input feature has
      higher values than the other one. Before use it make sure that
      input values are normalized and have similar scales.

    - Make sure that standard deviation in the same range as
      input features. Check ``std`` parameter description for
      more information.

    - The bigger training dataset the slower prediction.
      It's much more efficient for small datasets.

    {LazyLearningMixin.Notes}

    Methods
    -------
    {LazyLearningMixin.train}

    {BaseSkeleton.predict}

    {BaseSkeleton.fit}

    Examples
    --------
    >>> from sklearn import datasets, preprocessing
    >>> from sklearn.model_selection import train_test_split
    >>> from neupy import algorithms, estimators, environment
    >>>
    >>> environment.reproducible()
    >>>
    >>> dataset = datasets.load_diabetes()
    >>> x_train, x_test, y_train, y_test = train_test_split(
    ...     preprocessing.minmax_scale(dataset.data),
    ...     preprocessing.minmax_scale(dataset.target.reshape((-1, 1))),
    ...     test_size=0.3,
    ... )
    >>>
    >>> nw = algorithms.GRNN(std=0.1, verbose=False)
    >>> nw.train(x_train, y_train)
    >>>
    >>> y_predicted = nw.predict(x_test)
    >>> estimators.rmse(y_predicted, y_test)
    0.2381013391408185
    """
    std = BoundedProperty(default=0.1, minval=0)

    def train(self, input_train, target_train, copy=True):
        """
        Trains network. PNN doesn't actually train, it just stores
        input data and use it for prediction.

        Parameters
        ----------
        input_train : array-like (n_samples, n_features)

        target_train : array-like (n_samples,)
            Target variable should be vector or matrix
            with one feature column.

        copy : bool
            If value equal to ``True`` than input matrices will
            be copied. Defaults to ``True``.

        Raises
        ------
        ValueError
            In case if something is wrong with input data.
        """
        input_train = format_data(input_train, copy=copy)
        target_train = format_data(target_train, copy=copy)

        n_target_features = target_train.shape[1]
        if n_target_features != 1:
            raise ValueError("Target value must be one dimensional array")

        LazyLearningMixin.train(self, input_train, target_train)

    def predict(self, input_data):
        """
        Make a prediction from the input data.

        Parameters
        ----------
        input_data : array-like (n_samples, n_features)

        Raises
        ------
        ValueError
            In case if something is wrong with input data.

        Returns
        -------
        array-like (n_samples,)
        """
        if self.input_train is None:
            raise NotTrained("Cannot make a prediction. Network "
                             "hasn't been trained yet")

        input_data = format_data(input_data)

        input_data_size = input_data.shape[1]
        train_data_size = self.input_train.shape[1]

        if input_data_size != train_data_size:
            raise ValueError("Input data must contain {0} features, got "
                             "{1}".format(train_data_size, input_data_size))

        ratios = pdf_between_data(self.input_train, input_data, self.std)
        return (dot(self.target_train.T, ratios) / ratios.sum(axis=0)).T



import theano
import theano.tensor as T
import numpy as np

from neupy.utils import asfloat
from neupy.core.properties import NumberProperty
from .base import MinibatchGradientDescent


__all__ = ('Adagrad',)


class Adagrad(MinibatchGradientDescent):
    """
    Adagrad algorithm.

    Parameters
    ----------
    epsilon : float
        Value need to be greater than ``0``.
        Defaults to ``1e-5``.

    {MinibatchGradientDescent.Parameters}

    Attributes
    ----------
    {MinibatchGradientDescent.Attributes}

    Methods
    -------
    {MinibatchGradientDescent.Methods}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> mnet = algorithms.Adagrad((2, 3, 1))
    >>> mnet.train(x_train, y_train)
    """
    epsilon = NumberProperty(default=1e-5, minval=0)

    def init_param_updates(self, layer, parameter):
        step = self.variables.step

        parameter_shape = T.shape(parameter).eval()
        prev_mean_squred_grad = theano.shared(
            name="{}/prev-mean-squred-grad".format(parameter.name),
            value=asfloat(np.zeros(parameter_shape)),
        )

        gradient = T.grad(self.variables.error_func, wrt=parameter)

        mean_squred_grad = prev_mean_squred_grad + gradient ** 2
        parameter_delta = gradient * T.sqrt(mean_squred_grad + self.epsilon)

        return [
            (prev_mean_squred_grad, mean_squred_grad),
            (parameter, parameter - step * parameter_delta),
        ]



from neupy.core.config import Configurable
from neupy.algorithms.gd import SINGLE_STEP_UPDATE, MULTIPLE_STEP_UPDATE


__all__ = ('SingleStepConfigurable', 'MultipleStepConfigurable')


class SingleStepConfigurable(Configurable):
    """
    Configuration class for learning rate control algorithms.
    Works for algorithms that modify single learnig rate.

    Warns
    -----
    It works only with algorithms based on backpropagation.
    """
    addon_type = SINGLE_STEP_UPDATE


class MultipleStepConfigurable(Configurable):
    """
    Configuration class for learning rate control algorithms.
    Works for algorithms that modify multiple learnig rates for one
    neural network.

    Warns
    -----
    It works only with algorithms based on backpropagation.
    """
    addon_type = MULTIPLE_STEP_UPDATE



import random

import theano
import numpy as np


__all__ = ('sandbox', 'reproducible', 'speedup')


def sandbox():
    """
    Sandbox mode set up Theano configurations in the way that
    make compilation faster.
    """
    theano.config.optimizer = 'fast_compile'
    theano.config.allow_gc = False


def reproducible(seed=0):
    """
    Set up the same seed value for the NumPy and
    python random module to make your code reproducible.

    Parameters
    ----------
    seed : int
        Defaults to ``0``.
    """
    np.random.seed(seed)
    random.seed(seed)


def speedup():
    """
    Speed up Theano's computation.
    """
    theano.config.floatX = 'float32'
    theano.config.allow_gc = False



import theano
import theano.tensor as T
from theano.ifelse import ifelse
import numpy as np

from neupy.core.properties import (ChoiceProperty, NumberProperty,
                                   WithdrawProperty)
from neupy.algorithms.gd import StepSelectionBuiltIn
from neupy.algorithms.utils import parameter_values, setup_parameter_updates
from neupy.optimizations.wolfe import line_search
from neupy.layers.utils import count_parameters, iter_parameters
from neupy.utils import asfloat
from .base import GradientDescent


__all__ = ('QuasiNewton',)


def bfgs(inverse_hessian, weight_delta, gradient_delta, maxrho=1e4):
    ident_matrix = T.eye(inverse_hessian.shape[0])

    maxrho = asfloat(maxrho)
    rho = asfloat(1.) / gradient_delta.dot(weight_delta)

    rho = ifelse(
        T.isinf(rho),
        maxrho * T.sgn(rho),
        rho,
    )

    param1 = ident_matrix - T.outer(weight_delta, gradient_delta) * rho
    param2 = ident_matrix - T.outer(gradient_delta, weight_delta) * rho
    param3 = rho * T.outer(weight_delta, weight_delta)

    return param1.dot(inverse_hessian).dot(param2) + param3


def dfp(inverse_hessian, weight_delta, gradient_delta, maxnum=1e5):
    maxnum = asfloat(maxnum)
    quasi_dot_gradient = inverse_hessian.dot(gradient_delta)

    param1 = (
        T.outer(weight_delta, weight_delta)
    ) / (
        T.dot(gradient_delta, weight_delta)
    )
    param2_numerator = T.clip(
        T.outer(quasi_dot_gradient, gradient_delta) * inverse_hessian,
        -maxnum, maxnum
    )
    param2_denominator = gradient_delta.dot(quasi_dot_gradient)
    param2 = param2_numerator / param2_denominator

    return inverse_hessian + param1 - param2


def psb(inverse_hessian, weight_delta, gradient_delta, **options):
    gradient_delta_t = gradient_delta.T
    param = weight_delta - inverse_hessian.dot(gradient_delta)

    devider = (1. / T.dot(gradient_delta, gradient_delta))
    param1 = T.outer(param, gradient_delta) + T.outer(gradient_delta, param)
    param2 = (
        T.dot(gradient_delta, param) *
        T.outer(gradient_delta, gradient_delta_t)
    )

    return inverse_hessian + param1 * devider - param2 * devider ** 2


def sr1(inverse_hessian, weight_delta, gradient_delta, epsilon=1e-8):
    epsilon = asfloat(epsilon)
    param = weight_delta - inverse_hessian.dot(gradient_delta)
    denominator = T.dot(param, gradient_delta)

    return ifelse(
        T.lt(
            T.abs_(denominator),
            epsilon * param.norm(L=2) * gradient_delta.norm(L=2)
        ),
        inverse_hessian,
        inverse_hessian + T.outer(param, param) / denominator
    )


class QuasiNewton(StepSelectionBuiltIn, GradientDescent):
    """
    Quasi-Newton algorithm optimization.

    Parameters
    ----------
    update_function : {{'bfgs', 'dfp', 'psb', 'sr1'}}
        Update function. Defaults to ``bfgs``.

    h0_scale : float
        Default Hessian matrix is an identity matrix. The
        ``h0_scale`` parameter scales identity matrix.
        Defaults to ``1``.

    {GradientDescent.connection}

    {GradientDescent.error}

    {GradientDescent.show_epoch}

    {GradientDescent.shuffle_data}

    {GradientDescent.epoch_end_signal}

    {GradientDescent.train_end_signal}

    {GradientDescent.verbose}

    {GradientDescent.addons}

    Attributes
    ----------
    {GradientDescent.Attributes}

    Methods
    -------
    {GradientDescent.Methods}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> qnnet = algorithms.QuasiNewton(
    ...     (2, 3, 1),
    ...     update_function='bfgs'
    ... )
    >>> qnnet.train(x_train, y_train, epochs=10)

    See Also
    --------
    :network:`GradientDescent` : GradientDescent algorithm.
    """
    update_function = ChoiceProperty(
        default='bfgs',
        choices={
            'bfgs': bfgs,
            'dfp': dfp,
            'psb': psb,
            'sr1': sr1,
        }
    )
    h0_scale = NumberProperty(default=1, minval=0)

    step = WithdrawProperty()

    def init_variables(self):
        super(QuasiNewton, self).init_variables()
        n_params = count_parameters(self.connection)
        self.variables.update(
            inv_hessian=theano.shared(
                name='algo:quasi-newton/matrix:inv-hessian',
                value=asfloat(self.h0_scale * np.eye(int(n_params))),
            ),
            prev_params=theano.shared(
                name='algo:quasi-newton/vector:prev-params',
                value=asfloat(np.zeros(n_params)),
            ),
            prev_full_gradient=theano.shared(
                name='algo:quasi-newton/vector:prev-full-gradient',
                value=asfloat(np.zeros(n_params)),
            ),
        )

    def init_train_updates(self):
        network_inputs = self.variables.network_inputs
        network_output = self.variables.network_output
        inv_hessian = self.variables.inv_hessian
        prev_params = self.variables.prev_params
        prev_full_gradient = self.variables.prev_full_gradient

        params = parameter_values(self.connection)
        param_vector = T.concatenate([param.flatten() for param in params])

        gradients = T.grad(self.variables.error_func, wrt=params)
        full_gradient = T.concatenate([grad.flatten() for grad in gradients])

        new_inv_hessian = ifelse(
            T.eq(self.variables.epoch, 1),
            inv_hessian,
            self.update_function(inv_hessian,
                                 param_vector - prev_params,
                                 full_gradient - prev_full_gradient)
        )
        param_delta = -new_inv_hessian.dot(full_gradient)
        layers_and_parameters = list(iter_parameters(self.layers))

        def prediction(step):
            updated_params = param_vector + step * param_delta

            # This trick allow us to replace shared variables
            # with theano variables and get output from the network
            start_pos = 0
            for layer, attrname, param in layers_and_parameters:
                end_pos = start_pos + param.size
                updated_param_value = T.reshape(
                    updated_params[start_pos:end_pos],
                    param.shape
                )
                setattr(layer, attrname, updated_param_value)
                start_pos = end_pos

            output = self.connection.output(*network_inputs)

            # Restore previous parameters
            for layer, attrname, param in layers_and_parameters:
                setattr(layer, attrname, param)

            return output

        def phi(step):
            return self.error(network_output, prediction(step))

        def derphi(step):
            error_func = self.error(network_output, prediction(step))
            return T.grad(error_func, wrt=step)

        step = asfloat(line_search(phi, derphi))
        updated_params = param_vector + step * param_delta
        updates = setup_parameter_updates(params, updated_params)

        updates.extend([
            (inv_hessian, new_inv_hessian),
            (prev_params, param_vector),
            (prev_full_gradient, full_gradient),
        ])

        return updates



from neupy.core.properties import BoundedProperty
from neupy.utils import asfloat
from .base import WeightUpdateConfigurable


__all__ = ('WeightElimination',)


class WeightElimination(WeightUpdateConfigurable):
    """
    Weight Elimination algorithm penalizes large weights
    and limits the freedom in network. The algorithm is
    able to solve one of the possible problems of network
    overfitting.

    Parameters
    ----------
    decay_rate : float
        Controls the effect of penalties on the update
        network weights. Defaults to ``0.1``.

    zero_weight : float
        Second important parameter for weights penalization.
        Defaults to ``1``. Small value can make all weights
        close to zero. Big value will make less significant
        contribution in weights update. Which mean that with
        a bigger value of the ``zero_weight`` parameter network
        allows higher values for the weights.

    Warns
    -----
    {WeightUpdateConfigurable.Warns}

    Examples
    --------
    >>> from neupy import algorithms
    >>> bpnet = algorithms.GradientDescent(
    ...     (2, 4, 1),
    ...     step=0.1,
    ...     decay_rate=0.1,
    ...     addons=[algorithms.WeightElimination]
    ... )

    See Also
    --------
    :network:`WeightDecay` : Weight Decay penalty.

    Notes
    -----
    Before adding that regularization parameter carefully
    choose ``decay_rate`` and ``zero_weight`` parameters
    for the problem. Invalid parameters can make weight
    very close to the origin (all values become
    close to zero).

    References
    ----------
    [1] Weigend, A. S.; Rumelhart, D. E. & Huberman, B. A. (1991),
        Generalization by Weight-Elimination with Application to
        Forecasting, in Richard P. Lippmann; John E. Moody & David S.
        Touretzky, ed., Advances in Neural Information Processing
        Systems, San Francisco, CA: Morgan Kaufmann, pp. 875--882 .
    """
    decay_rate = BoundedProperty(default=0.1, minval=0)
    zero_weight = BoundedProperty(default=1, minval=0)

    def init_param_updates(self, layer, parameter):
        updates = super(WeightElimination, self).init_param_updates(
            layer, parameter
        )

        step = self.variables.step
        decay_koef = asfloat(self.decay_rate * step)
        zero_weight_square = asfloat(self.zero_weight ** 2)

        updates_mapper = dict(updates)
        updates_mapper[parameter] -= decay_koef * (
            (2 * parameter / zero_weight_square) / (
                1 + (parameter ** 2) / zero_weight_square
            ) ** 2
        )

        return list(updates_mapper.items())



import theano
import theano.tensor as T
import numpy as np
import matplotlib.pyplot as plt
from scipy.ndimage.filters import gaussian_filter

from neupy.exceptions import InvalidConnection
from neupy.layers.utils import create_input_variable


__all__ = ('saliency_map', 'compile_saliency_map')


def compile_saliency_map(connection):
    """
    Compile Theano function that returns saliency map.

    Parameters
    ----------
    connection : connection
    """
    x = create_input_variable(connection.input_shape,
                              name='plots:saliency-map/var:input')

    with connection.disable_training_state():
        prediction = connection.output(x)

    output_class = T.argmax(prediction)
    saliency = T.grad(T.max(prediction), x)

    return theano.function([x], [saliency, output_class])


def saliency_map(connection, image, mode='heatmap', sigma=8,
                 ax=None, show=True, **kwargs):
    """
    Saliency Map plot.

    Parameters
    ----------
    connection : network, connection
        Network based on which will be computed saliency map.

    image : 3D array-like tensor
        Image based on which will be computed saliency map.

    mode : {``raw``, ``heatmap``}
        - ``raw``
          Visualize raw gradient. White color on the plot
          defines high gradient values.

        - ``heatmap``
          Applies gaussian filter to the gradient and visualize
          as a heatmap plot.

        Defaults to ``heatmap``.

    sigma : float
        Standard deviation for kernel in Gaussian filter.
        It is used only when ``mode='heatmap'``. Defaults to ``8``.

    ax : object or None
        Matplotlib axis object. ``None`` values means that axis equal
        to the current axes instance (the same as ``ax = plt.gca()``).
        Defaults to ``None``.

    show : bool
        If parameter is equal to ``True`` then plot will be
        displayed. Defaults to ``True``.

    **kwargs
        Arguments for ``plt.imshow`` function.

    Returns
    -------
    object
        Matplotlib axis instance.

    Examples
    --------
    >>> from neupy import layers, plots
    >>>
    >>> network = layers.join(
    ...     layers.Input((3, 28, 28)),
    ...     layers.Convolution((32, 3, 3)) > layers.Relu(),
    ...     layers.Reshape(),
    ...     layers.Softmax(10),
    ... )
    >>>
    >>> dog_image = load_dog_image()
    >>> plots.saliency_map(network, dog_image)
    """
    if image.ndim == 3:
        image = np.expand_dims(image, axis=0)

    if image.ndim != 4:
        raise ValueError("Invalid image shape. Image expected to be 3D, "
                         "got {}D image".format(image.ndim))

    valid_modes = ('raw', 'heatmap')
    if mode not in valid_modes:
        raise ValueError("{!r} is invalid value for mode argument. Valid "
                         "mode values are: {!r}".format(mode, valid_modes))

    if len(connection.output_layers) != 1:
        raise InvalidConnection(
            "Cannot build saliency map for connection that has more than "
            "one output layer. Output layer can be specified explicitly. "
            "For instance, use network.end('layer-name') with specified "
            "output layer name.")

    if len(connection.input_layers) != 1:
        raise InvalidConnection(
            "Cannot build saliency map for connection that has more than "
            "one input layer. Output layer can be specified explicitly. "
            "For instance, use network.start('layer-name') with specified "
            "output layer name.")

    if len(connection.input_shape) != 3:
        raise InvalidConnection("Input layer has invalid input shape")

    if ax is None:
        ax = plt.gca()

    saliency_and_output = compile_saliency_map(connection)
    saliency, output = saliency_and_output(image)

    saliency = saliency[0].transpose((1, 2, 0))
    saliency = saliency.max(axis=2)

    if mode == 'heatmap':
        saliency = gaussian_filter(saliency, sigma=sigma)

    elif mode == 'raw':
        kwargs.setdefault('cmap', 'gray')

    ax.set_title('Predicted output #{} (0-based indeces)'.format(output))
    ax.imshow(saliency, **kwargs)

    if show:
        plt.show()

    return ax



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals

import os
import sys
import importlib
from contextlib import contextmanager

import tableprint
from tableprint.style import STYLES, TableStyle, LineStyle

from .config import Configurable
from .properties import BaseProperty
from . import terminal


__all__ = ('Verbose',)


# Customized style
STYLES['round'] = TableStyle(
        top=LineStyle('--', '-', '---', '--'),
        below_header=LineStyle('--', '-', '---', '--'),
        bottom=LineStyle('--', '-', '---', '--'),
        row=LineStyle('| ', '', ' | ', ' |'),
)


def terminal_echo(enabled, file_descriptor=sys.stdin):
    """
    Enable/disable echo in the terminal.

    Parameters
    ----------
    enabled : bool
        The `False` value means that you willn't be able to make
        input in terminal.
    file_descriptor : object
        File descriptor that you would like to enable or disable.
        Defaults to ``sys.stdin``.
    """
    is_windows = (os.name == 'nt')

    if is_windows:
        return

    termios = importlib.import_module('termios')

    try:
        attributes = termios.tcgetattr(file_descriptor)
        lflag = attributes[3]

        if enabled:
            lflag |= termios.ECHO
        else:
            lflag &= ~termios.ECHO

        attributes[3] = lflag
        termios.tcsetattr(file_descriptor, termios.TCSANOW, attributes)

    except Exception:
        # This feature is not very important, but any error in
        # this function can crash program, so it's easier to
        # ignore it in case if something went wrong
        pass


class TerminalLogger(object):
    """
    Customized logging class that replace standard logging
    functionality.

    Attributes
    ----------
    enable : bool
        Enable/disable logging output. Defaults to ``True``.
    template : str
        Terminal output message template. Defaults
        to ``"[{name}] {text}"``.
    stdout : object
        Writes output in terminal. Defaults to ``sys.stdout``.
    """

    colors = {
        'gray': terminal.gray,
        'green': terminal.green,
        'red': terminal.red,
        'white': terminal.white,
    }

    def __init__(self, enable=True):
        self.enable = enable
        self.template = "[{tag}] {text}"
        self.stdout = sys.stdout

    def write(self, text):
        """
        Method writes text in terminal if logging is enable.

        Parameters
        ----------
        text : str
        """
        if self.enable:
            self.stdout.write(text)
            self.stdout.write('\n')

    def newline(self):
        """
        Just writes an empty line.
        """
        self.write('\r')

    def message(self, tag, text, color='green'):
        """
        Methods writes message in terminal using specific template.
        Each row should have tag and text. Tag identifies message
        category and text information reletad to this category.

        Parameters
        ----------
        name : str
        text : str
        color : {{'green', 'gray', 'red', 'white'}}
            Property that color text defined as ``tag`` parameter.
            Defaults to ``green``.
        """
        if color not in self.colors:
            available_colors = ', '.join(self.colors.keys())
            raise ValueError("Invalid color `{}`. Available colors: {}"
                             "".format(color, available_colors))

        colorizer = self.colors[color]
        formated_tag = colorizer(tag.upper())
        message = self.template.format(tag=formated_tag, text=text)
        self.write(message)

    def title(self, text):
        """
        Method write text as a title message. Text will be displayed
        using bold and underline text styles. Also there will be empty
        lines before and after the message.

        Parameters
        ----------
        text : str
        """
        bold_text = terminal.bold(text)
        message = "\n{text}\n".format(text=terminal.underline(bold_text))
        self.write(message)

    def error(self, text):
        """
        Method writes messages that related to error type.
        Text will be displayed as message with ``tag`` parameter equal
        to ``'ERROR'``. Color will be red.

        Parameters
        ----------
        text : str
        """
        self.message('ERROR', text, color='red')

    def warning(self, text):
        """
        Method writes messages that related to warning type.
        Text will be displayed as message with ``tag`` parameter equal
        to ``'WARN'``. Color will be red.

        Parameters
        ----------
        text : str
        """
        self.message('WARN', text, color='red')

    @contextmanager
    def disable_user_input(self):
        """
        Context manager helps ignore user input in
        terminal for UNIX.
        """
        if self.enable:
            try:
                terminal_echo(enabled=False)
                yield
            finally:
                terminal_echo(enabled=True)
        else:
            yield

    def __reduce__(self):
        return (self.__class__, (self.enable,))

    def table_header(self, header, *args, **kwargs):
        self.write(tableprint.header(header, *args, **kwargs))

    def table_row(self, row, *args, **kwargs):
        self.write(tableprint.row(row, *args, **kwargs))

    def table_bottom(self, n_columns, *args, **kwargs):
        self.write(tableprint.bottom(n_columns, *args, **kwargs))

    def table(self, data, headers, **kwargs):
        if not self.enable:
            return

        widths = [len(value) for value in headers]
        stringified_data = []

        for row_values in data:
            stringified_data.append([str(v) for v in row_values])

            for i, cell_value in enumerate(row_values):
                widths[i] = max(len(str(cell_value)), widths[i])

        kwargs['width'] = widths
        kwargs['out'] = self.stdout
        tableprint.table(stringified_data, headers, **kwargs)


class VerboseProperty(BaseProperty):
    """
    Property that synchronize updates with ``enable`` attribute in
    logging instance.

    Parameters
    ----------
    {BaseProperty.default}
    {BaseProperty.required}
    """
    expected_type = bool

    def __set__(self, instance, value):
        instance.logs.enable = value
        return super(VerboseProperty, self).__set__(instance, value)


class Verbose(Configurable):
    """
    Class that controls NeuPy logging.

    Parameters
    ----------
    verbose : bool
        Property controls verbose output interminal. ``True`` enables
        informative output in the terminal and ``False`` -
        disable it. Defaults to ``False``.

    Attributes
    ----------
    logs : TerminalLogger
        ``TerminalLogger`` instance.
    """
    verbose = VerboseProperty(default=False)

    def __init__(self, **options):
        self.logs = TerminalLogger()
        self.logs.enable = self.verbose
        super(Verbose, self).__init__(**options)



import numpy as np
import theano.tensor as T


from neupy.layers.utils import iter_parameters


__all__ = ('shuffle', 'parameter_values', 'iter_until_converge',
           'setup_parameter_updates')


def parameter_values(connection):
    """
    List of all trainable parameters in the network.

    Parameters
    ----------
    connection : layer, connection

    Yields
    ------
    Theano shared variable
        Network's trainable parameter.
    """
    parameters = []

    for _, _, parameter in iter_parameters(connection):
        parameters.append(parameter)

    return parameters


def setup_parameter_updates(parameters, parameter_update_vector):
    """
    Creates update rules for list of parameters from one vector.
    Function is useful in Conjugate Gradient or
    Levenberg-Marquardt optimization algorithms

    Parameters
    ----------
    parameters : list
        List of parameters.

    parameter_update_vector : Theano varible
        Vector that contains updates for all parameters.

    Returns
    -------
    list
        List of updates separeted for each parameter.
    """
    updates = []
    start_position = 0

    for parameter in parameters:
        end_position = start_position + parameter.size

        new_parameter = T.reshape(
            parameter_update_vector[start_position:end_position],
            parameter.shape
        )
        updates.append((parameter, new_parameter))

        start_position = end_position

    return updates


def iter_until_converge(network, epsilon, max_epochs):
    """
    Train network until error converged or maximum number of
    epochs has been reached.

    Parameters
    ----------
    network : BaseNetwork instance

    epsilon : float
        Interrupt training in case if different absolute
        between two previous errors is less than specified
        epsilon value.

    max_epochs : int
        Maximum number of epochs to train.
    """
    logs = network.logs

    # Trigger first iteration and store first error term
    yield network.last_epoch

    previous_error = error_delta = network.errors.last()
    epoch = network.last_epoch

    while error_delta > epsilon:
        epoch = epoch + 1
        network.last_epoch += 1

        yield epoch

        last_error = network.errors.last()
        error_delta = abs(last_error - previous_error)
        previous_error = last_error

        if epoch >= max_epochs and error_delta > epsilon:
            logs.message("TRAIN", "Epoch #{} interrupted. Network didn't "
                                  "converge after {} iterations"
                                  "".format(epoch, max_epochs))
            return

    if np.isnan(error_delta) or np.isinf(error_delta):
        logs.message("TRAIN", "Epoch #{} interrupted. Network error value is "
                              "NaN or Inf.".format(epoch))
    else:
        logs.message("TRAIN", "Epoch #{} interrupted. Network converged."
                              "".format(epoch))


def shuffle(*arrays):
    """
    Randomly shuffle rows in the arrays qithout breaking
    associations between rows in different arrays.

    Parameters
    ----------
    *arrays
        Arrays that should be shuffled.

    Returns
    -------
    list
        List of arrays that contain shuffeled input data.
    """
    filtered_arrays = tuple(array for array in arrays if array is not None)

    if not filtered_arrays:
        return arrays

    first = filtered_arrays[0]
    n_samples = first.shape[0]

    if any(n_samples != array.shape[0] for array in filtered_arrays):
        array_shapes = [array.shape for array in filtered_arrays]
        raise ValueError("Cannot shuffle matrices. All matrices should "
                         "have the same number of rows. Input shapes are: {}"
                         "".format(array_shapes))

    indices = np.arange(n_samples)
    np.random.shuffle(indices)

    arrays = list(arrays)
    for i, array in enumerate(arrays):
        if array is not None:
            arrays[i] = array[indices]

    if len(arrays) == 1:
        return arrays[0]

    return tuple(arrays)



import numpy as np

from neupy.utils import format_data
from neupy.exceptions import NotTrained
from neupy.core.properties import BoundedProperty
from neupy.algorithms.base import BaseNetwork
from neupy.algorithms.gd.base import MinibatchTrainingMixin
from .learning import LazyLearningMixin
from .utils import pdf_between_data


__all__ = ('PNN',)


class PNN(BaseNetwork, LazyLearningMixin, MinibatchTrainingMixin):
    """
    Probabilistic Neural Network (PNN). Network applies only to
    the classification problems.

    Notes
    -----
    - PNN Network is sensitive for cases when one input feature
      has higher values than the other one. Before use it make
      sure that input values are normalized and have similar scales.

    - Make sure that standard deviation in the same range as
      input features. Check ``std`` parameter description for
      more information.

    - The bigger training dataset the slower prediction.
      It's much more efficient for small datasets.

    {LazyLearningMixin.Notes}

    Parameters
    ----------
    std : float
        Standard deviation for the Probability Density Function (PDF).
        Defaults to ``0.1``. If your input features have high values
        than standard deviation should also be high. For instance,
        if input features from range ``[0, 20]`` that standard
        deviation should be also a big value like ``10`` or ``15``.
        Small values will lead to bad prediction.

    {MinibatchTrainingMixin.batch_size}

    {BaseNetwork.verbose}

    Methods
    -------
    {LazyLearningMixin.train}

        The ``target_train`` argument should be a vector or
        matrix with one feature column.

    {BaseSkeleton.predict}

    predict_proba(input_data)
        Predict probabilities for each class.

    {BaseSkeleton.fit}

    Examples
    --------
    >>> import numpy as np
    >>>
    >>> from sklearn import datasets, metrics
    >>> from sklearn.model_selection import train_test_split
    >>> from neupy import algorithms, environment
    >>>
    >>> environment.reproducible()
    >>>
    >>> dataset = datasets.load_digits()
    >>> x_train, x_test, y_train, y_test = train_test_split(
    ...     dataset.data, dataset.target, test_size=0.3
    ... )
    >>>
    >>> pnn = algorithms.PNN(std=10, verbose=False)
    >>> pnn.train(x_train, y_train)
    >>>
    >>> y_predicted = pnn.predict(x_test)
    >>> metrics.accuracy_score(y_test, y_predicted)
    0.98888888888888893
    """
    std = BoundedProperty(default=0.1, minval=0)

    def __init__(self, **options):
        super(PNN, self).__init__(**options)
        self.classes = None

    def train(self, input_train, target_train, copy=True):
        """
        Trains network. PNN doesn't actually train, it just stores
        input data and use it for prediction.

        Parameters
        ----------
        input_train : array-like (n_samples, n_features)

        target_train : array-like (n_samples,)
            Target variable should be vector or matrix
            with one feature column.

        copy : bool
            If value equal to ``True`` than input matrices will
            be copied. Defaults to ``True``.

        Raises
        ------
        ValueError
            In case if something is wrong with input data.
        """
        input_train = format_data(input_train, copy=copy)
        target_train = format_data(target_train, copy=copy, make_float=False)

        LazyLearningMixin.train(self, input_train, target_train)

        n_target_features = target_train.shape[1]
        if n_target_features != 1:
            raise ValueError("Target value should be a vector or a "
                             "matrix with one column")

        classes = self.classes = np.unique(target_train)
        n_classes = classes.size
        n_samples = input_train.shape[0]

        class_ratios = self.class_ratios = np.zeros(n_classes)
        row_comb_matrix = self.row_comb_matrix = np.zeros(
            (n_classes, n_samples))

        for i, class_name in enumerate(classes):
            class_name = classes[i]
            class_val_positions = (target_train == class_name)
            row_comb_matrix[i, class_val_positions.ravel()] = 1
            class_ratios[i] = np.sum(class_val_positions)

    def predict_proba(self, input_data):
        """
        Predict probabilities for each class.

        Parameters
        ----------
        input_data : array-like (n_samples, n_features)

        Returns
        -------
        array-like (n_samples, n_classes)
        """
        outputs = self.apply_batches(
            function=self.predict_raw,
            input_data=format_data(input_data),

            description='Prediction batches',
            show_progressbar=True,
            show_error_output=False,
        )
        raw_output = np.concatenate(outputs, axis=1)

        total_output_sum = raw_output.sum(axis=0).reshape((-1, 1))
        return raw_output.T / total_output_sum

    def predict_raw(self, input_data):
        """
        Raw prediction.

        Parameters
        ----------
        input_data : array-like (n_samples, n_features)

        Raises
        ------
        NotTrained
            If network hasn't been trained.

        ValueError
            In case if something is wrong with input data.

        Returns
        -------
        array-like (n_samples, n_classes)
        """
        if self.classes is None:
            raise NotTrained("Cannot make a prediction. Network "
                             "hasn't been trained yet")

        input_data_size = input_data.shape[1]
        train_data_size = self.input_train.shape[1]

        if input_data_size != train_data_size:
            raise ValueError("Input data must contain {0} features, got "
                             "{1}".format(train_data_size, input_data_size))

        class_ratios = self.class_ratios.reshape((-1, 1))
        pdf_outputs = pdf_between_data(self.input_train, input_data, self.std)

        return np.dot(self.row_comb_matrix, pdf_outputs) / class_ratios

    def predict(self, input_data):
        """
        Predicts class from the input data.

        Parameters
        ----------
        input_data : array-like (n_samples, n_features)

        Returns
        -------
        array-like (n_samples,)
        """
        outputs = self.apply_batches(
            function=self.predict_raw,
            input_data=format_data(input_data),

            description='Prediction batches',
            show_progressbar=True,
            show_error_output=False,
        )
        raw_output = np.concatenate(outputs, axis=1)
        return self.classes[raw_output.argmax(axis=0)]



import collections

import six
import theano.tensor as T

from neupy.utils import as_tuple
from neupy.exceptions import LayerConnectionError
from neupy.core.properties import TypedListProperty, Property
from .base import ParameterBasedLayer


__all__ = ('Convolution',)


class StrideProperty(TypedListProperty):
    """
    Stride property.

    Parameters
    ----------
    {BaseProperty.Parameters}
    """
    expected_type = (list, tuple, int)

    def __init__(self, *args, **kwargs):
        kwargs['element_type'] = int
        super(StrideProperty, self).__init__(*args, **kwargs)

    def __set__(self, instance, value):
        if isinstance(value, collections.Iterable) and len(value) == 1:
            value = (value[0], 1)

        if isinstance(value, int):
            value = (value, value)

        super(StrideProperty, self).__set__(instance, value)

    def validate(self, value):
        super(StrideProperty, self).validate(value)

        if len(value) > 2:
            raise ValueError("Stide can have only one or two elements "
                             "in the list. Got {}".format(len(value)))

        if any(element <= 0 for element in value):
            raise ValueError("Stride size should contain only values greater "
                             "than zero")


class BorderModeProperty(Property):
    """
    Border mode property identifies border for the
    convolution operation.

    Parameters
    ----------
    {Property.Parameters}
    """
    expected_type = (six.string_types, int, tuple)
    valid_string_choices = ('valid', 'full', 'half')

    def validate(self, value):
        super(BorderModeProperty, self).validate(value)

        if isinstance(value, tuple) and len(value) != 2:
            raise ValueError(
                "Border mode property suppose to get a tuple that "
                "contains two elements, got {} elements"
                "".format(len(value))
            )

        is_invalid_string = (
            isinstance(value, six.string_types) and
            value not in self.valid_string_choices
        )
        if is_invalid_string:
            valid_choices = ', '.join(self.valid_string_choices)
            raise ValueError("`{}` is invalid string value. Available: {}"
                             "".format(value, valid_choices))

        if isinstance(value, int) and value < 0:
            raise ValueError("Integer border mode value needs to be "
                             "greater or equal to zero, got {}".format(value))

        if isinstance(value, tuple) and any(element < 0 for element in value):
            raise ValueError("Tuple border mode value needs to contain "
                             "only elements that greater or equal to zero, "
                             "got {}".format(value))


def conv_output_shape(dimension_size, filter_size, padding, stride):
    """
    Computes convolution's output shape.

    Parameters
    ----------
    dimension_size : int
        Size of the dimension. Typically it's image's
        weight or height.

    filter_size : int
        Size of the convolution filter.

    padding : {``valid``, ``full``, ``half``} or int
        Type or size of the zero-padding.

    stride : int
        Stride size.

    Returns
    -------
    int
        Dimension size after applying convolution
        operation with specified configurations.
    """
    if dimension_size is None:
        return None

    if not isinstance(stride, int):
        raise ValueError("Stride needs to be an integer, got {} (value {!r})"
                         "".format(type(stride), stride))

    if not isinstance(filter_size, int):
        raise ValueError("Filter size needs to be an integer, got {} "
                         "(value {!r})".format(type(filter_size),
                                               filter_size))

    if padding == 'valid':
        output_size = dimension_size - filter_size + 1

    elif padding == 'half':
        output_size = dimension_size + 2 * (filter_size // 2) - filter_size + 1

    elif padding == 'full':
        output_size = dimension_size + filter_size - 1

    elif isinstance(padding, int):
        output_size = dimension_size + 2 * padding - filter_size + 1

    else:
        raise ValueError("`{!r}` is unknown convolution's border mode value"
                         "".format(padding))

    return (output_size + stride - 1) // stride


class Convolution(ParameterBasedLayer):
    """
    Convolutional layer.

    Parameters
    ----------
    size : tuple of int
        Filter shape. In should be defined as a tuple with three integers
        ``(output channels, filter rows, filter columns)``.

    padding : {{``valid``, ``full``, ``half``}} or int or tuple with 2 int
        Convolution border mode. Check Theano's ``nnet.conv2d`` doc.
        Defaults to ``valid``.

    stride : tuple with 1 or 2 integers or integer.
        Stride size. Defaults to ``(1, 1)``

    {ParameterBasedLayer.weight}

    {ParameterBasedLayer.bias}

    {BaseLayer.Parameters}

    Examples
    --------
    2D Convolution

    >>> from neupy import layers
    >>>
    >>> layers.join(
    ...     layers.Input((3, 28, 28)),
    ...     layers.Convolution((16, 3, 3)),
    ... )

    1D Convolution

    >>> from neupy import layers
    >>>
    >>> layers.join(
    ...     layers.Input((10, 30)),
    ...     layers.Reshape((10, 30, 1)),
    ...     layers.Convolution((16, 3, 1)),
    ... )

    Methods
    -------
    {ParameterBasedLayer.Methods}

    Attributes
    ----------
    {ParameterBasedLayer.Attributes}
    """
    size = TypedListProperty(required=True, element_type=int)
    padding = BorderModeProperty(default='valid')
    stride = StrideProperty(default=(1, 1))

    def validate(self, input_shape):
        if len(input_shape) != 3:
            raise LayerConnectionError(
                "Convolutional layer expects an input with 3 "
                "dimensions, got {} with shape {}"
                "".format(len(input_shape), input_shape))

    @property
    def output_shape(self):
        if self.input_shape is None:
            return None

        padding = self.padding
        n_kernels = self.size[0]
        rows, cols = self.input_shape[-2:]
        row_filter_size, col_filter_size = self.size[-2:]
        row_stride, col_stride = self.stride

        if isinstance(padding, tuple):
            row_padding, col_padding = padding[-2:]
        else:
            row_padding, col_padding = padding, padding

        output_rows = conv_output_shape(rows, row_filter_size,
                                        row_padding, row_stride)
        output_cols = conv_output_shape(cols, col_filter_size,
                                        col_padding, col_stride)
        return (n_kernels, output_rows, output_cols)

    @property
    def weight_shape(self):
        n_channels = self.input_shape[0]
        n_filters, n_rows, n_cols = self.size
        return (n_filters, n_channels, n_rows, n_cols)

    @property
    def bias_shape(self):
        return as_tuple(self.size[0])

    def output(self, input_value):
        output = T.nnet.conv2d(
            input_value, self.weight,
            input_shape=as_tuple(None, self.input_shape),
            filter_shape=self.weight_shape,
            border_mode=self.padding,
            subsample=self.stride)

        if self.bias is not None:
            bias = T.reshape(self.bias, (1, -1, 1, 1))
            output += bias

        return output



from neupy import layers
from neupy.utils import all_equal
from neupy.layers.utils import extract_connection


__all__ = ('mixture_of_experts',)


def check_if_connection_is_valid(connection, index):
    if len(connection.input_layers) > 1:
        raise ValueError(
            "Network #{} (0-based indeces) has more than one input "
            "layer. Input layers: {!r}"
            "".format(index, connection.input_layers))

    if len(connection.output_layers) > 1:
        raise ValueError(
            "Network #{} (0-based indeces) has more than one output "
            "layer. Output layers: {!r}"
            "".format(index, connection.output_layers))

    if len(connection.input_shape) != 1:
        raise ValueError(
            "Network #{} (0-based indeces) should receive vector as "
            "an input. Input layer shape: {!r}"
            "".format(index, connection.input_shape))


def check_if_connections_compatible(connections):
    input_shapes = []
    output_shapes = []

    for i, connection in enumerate(connections):
        input_shapes.append(connection.input_shape)
        output_shapes.append(connection.output_shape)

    if not all_equal(input_shapes):
        raise ValueError("Networks have different input shapes: {}"
                         "".format(input_shapes))

    if not all_equal(output_shapes):
        raise ValueError("Networks have different output shapes: {}"
                         "".format(output_shapes))


def check_if_gating_layer_valid(gating_layer, n_layers_to_combine):
    if not isinstance(gating_layer, layers.BaseLayer):
        raise ValueError(
            "Invalid type for gating layer. Type: {}"
            "".format(type(gating_layer)))

    output_shape = gating_layer.output_shape[0]
    if output_shape != n_layers_to_combine:
        raise ValueError(
            "Gating layer has invalid number of outputs. Expected {}, got {}"
            "".format(output_shape, n_layers_to_combine))


def mixture_of_experts(networks, gating_layer=None):
    """
    Generates mixture of experts architecture from the set of
    networks that has the same input and output shapes.

    Mixture of experts learns to how to mix results from different
    networks in order to get better performances. It adds gating layer
    that using input data tries to figure out which of the networks
    will make better contribution to the final result. The final result
    mixes from all networks using different weights. The higher the weight
    the larger contribution from the individual layer.

    Parameters
    ----------
    networks : list of connections or networks
        These networks will be combine into mixture of experts.
        Every network should have single 1D input layer and
        single output layer. Another restriction is that all networks
        should expect the same input and output layers.

    gating_layer : None or layer
        In case if value equal to `None` that the following layer
        will be created.

        .. code-block:: python

            gating_layer = layers.Softmax(len(networks))

        Output from the gating layer should be 1D and equal to
        the number of networks.

    Raises
    ------
    ValueError
        In case if there is some problem with input networks
        or custom gating layer.

    Returns
    -------
    connection
        Mixture of experts network that combine all networks into
        single one and adds gating layer to it.

    Examples
    --------
    >>> from neupy import layers, algorithms, architectures
    >>>
    >>> network = architectures.mixture_of_experts([
    ...     layers.join(
    ...         layers.Input(10),
    ...         layers.Relu(5),
    ...     ),
    ...     layers.join(
    ...         layers.Input(10),
    ...         layers.Relu(33),
    ...         layers.Relu(5),
    ...     ),
    ...     layers.join(
    ...         layers.Input(10),
    ...         layers.Relu(12),
    ...         layers.Relu(25),
    ...         layers.Relu(5),
    ...     ),
    ... ])
    >>> network
    10 -> [... 12 layers ...] -> 5
    >>>
    >>> gdnet = algorithms.Momentum(network, step=0.1)
    """
    if not isinstance(networks, (list, tuple)):
        raise ValueError("Networks should be specified as a list")

    connections = []
    for index, network in enumerate(networks):
        connection = extract_connection(network)
        check_if_connection_is_valid(connection, index)
        connections.append(connection)

    check_if_connections_compatible(connections)

    first_connection = connections[0]
    n_features = first_connection.input_shape[0]
    n_layers_to_combine = len(connections)

    if gating_layer is None:
        gating_layer = layers.Softmax(n_layers_to_combine)

    check_if_gating_layer_valid(gating_layer, n_layers_to_combine)

    return layers.join(
        layers.Input(n_features),
        # Note: Gating network should be specified
        # as a first parameter.
        [gating_layer] + connections,
        layers.GatedAverage())



from __future__ import division

import theano
import numpy as np
from sklearn import datasets, metrics
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import OneHotEncoder
from neupy import algorithms, layers, environment


environment.reproducible()
theano.config.floatX = 'float32'
theano.config.allow_gc = False

mnist = datasets.fetch_mldata('MNIST original')
data = mnist.data / 255.

target_scaler = OneHotEncoder()
target = mnist.target.reshape((-1, 1))
target = target_scaler.fit_transform(target).todense()

n_labeled = 1000
n_samples = len(data)
n_unlabeled = n_samples - n_labeled

x_labeled, x_unlabeled, y_labeled, y_unlabeled = train_test_split(
    data.astype(np.float32),
    target.astype(np.float32),
    test_size=(1 - n_labeled / n_samples)
)

x_labeled_4d = x_labeled.reshape((n_labeled, 1, 28, 28))
x_unlabeled_4d = x_unlabeled.reshape((n_unlabeled, 1, 28, 28))

encoder = layers.join(
    layers.Input((1, 28, 28)),

    layers.Convolution((16, 3, 3)) > layers.Relu(),
    layers.Convolution((16, 3, 3)) > layers.Relu(),
    layers.MaxPooling((2, 2)),

    layers.Convolution((32, 3, 3)) > layers.Relu(),
    layers.MaxPooling((2, 2)),

    layers.Reshape(),

    layers.Relu(256),
    layers.Relu(128),
)

decoder = layers.join(
    layers.Relu(256),
    layers.Relu(32 * 5 * 5),

    layers.Reshape((32, 5, 5)),

    layers.Upscale((2, 2)),
    layers.Convolution((16, 3, 3), padding='full') > layers.Relu(),

    layers.Upscale((2, 2)),
    layers.Convolution((16, 3, 3), padding='full') > layers.Relu(),
    layers.Convolution((1, 3, 3), padding='full') > layers.Sigmoid(),

    layers.Reshape(),
)

conv_autoencoder = algorithms.Momentum(
    connection=encoder > decoder,
    verbose=True,
    step=0.1,
    momentum=0.99,
    shuffle_data=True,
    batch_size=64,
    error='binary_crossentropy',
)
conv_autoencoder.architecture()
conv_autoencoder.train(x_unlabeled_4d, x_unlabeled,
                       x_labeled_4d, x_labeled, epochs=10)

x_labeled_encoded = encoder.output(x_labeled_4d).eval()
x_unlabeled_encoded = encoder.output(x_unlabeled_4d).eval()

classifier_network = layers.join(
    layers.PRelu(512),
    layers.Dropout(0.25),
    layers.Softmax(10),
)

encoder_classifier = algorithms.Adadelta(
    layers.Input(encoder.output_shape) > classifier_network,
    verbose=True,
    step=0.05,
    shuffle_data=True,
    batch_size=64,
    error='categorical_crossentropy',
)
encoder_classifier.architecture()
encoder_classifier.train(x_labeled_encoded, y_labeled,
                         x_unlabeled_encoded, y_unlabeled, epochs=100)

classifier = algorithms.MinibatchGradientDescent(
    encoder > classifier_network,
    verbose=True,
    step=0.01,
    shuffle_data=True,
    batch_size=64,
    error='categorical_crossentropy',
)
classifier.architecture()
classifier.train(x_labeled_4d, y_labeled, epochs=100)

unlabeled_predicted = classifier.predict(x_unlabeled_4d).argmax(axis=1)
y_unlabeled_classes = np.asarray(y_unlabeled).argmax(axis=1)

print(metrics.classification_report(y_unlabeled_classes, unlabeled_predicted))
score = metrics.accuracy_score(y_unlabeled_classes, unlabeled_predicted)
print("Validation accuracy: {:.2%}".format(score))



from __future__ import division

import math

import six
import theano
import theano.tensor as T
import numpy as np
import progressbar

from neupy.core.config import Configurable
from neupy.core.properties import Property, BoundedProperty
from neupy.utils import as_tuple
from neupy.algorithms.constructor import ConstructibleNetwork
from neupy.algorithms.gd import addon_types


__all__ = ('GradientDescent', 'MinibatchGradientDescent')


class GradientDescent(ConstructibleNetwork):
    """
    Gradient descent algorithm.

    Parameters
    ----------
    {ConstructibleNetwork.Parameters}

    addons : list or None
        The list of addon algortihms. ``None`` by default.
        If this option is not empty it will generate new class which
        will inherit all from this list. Support two types of
        addon algorithms: weight update and step update.

    Attributes
    ----------
    {ConstructibleNetwork.Attributes}

    Methods
    -------
    {ConstructibleNetwork.Methods}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy.algorithms import GradientDescent
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> bpnet = GradientDescent((2, 3, 1), verbose=False, step=0.1)
    >>> bpnet.train(x_train, y_train)
    """
    supported_addon_types = addon_types.keys()

    addons = Property(default=None, expected_type=list)

    # TODO: The arguments that have default value equal to `None`
    # are useful only in case if we need to save network in the
    # file. This solution looks bad and I need to redesign it later.
    def __new__(cls, connection=None, options=None, floatX=None, **kwargs):
        # Argument `options` is a simple hack for the `__reduce__` method.
        # `__reduce__` can't retore class with keyword arguments and
        # it will put them as `dict` argument in the `options` and method
        # will translate it to kwargs. The same hack is in the
        # `__init__` method.

        if options is None:
            options = kwargs

        if floatX is not None:
            theano.config.floatX = floatX

        addons = options.get('addons')

        if not addons:
            cls.main_class = cls
            return super(GradientDescent, cls).__new__(cls)

        identified_types = []
        for addon_class in addons:
            opt_class_type = getattr(addon_class, 'addon_type',  None)

            if opt_class_type not in cls.supported_addon_types:
                opt_class_name = addon_class.__name__
                supported_opts = ', '.join(addon_types.values())
                raise ValueError(
                    "Invalid add-on class '{}'. Class supports only "
                    "{}".format(opt_class_name, supported_opts)
                )

            if opt_class_type in identified_types:
                raise ValueError(
                    "There can be only one add-on class with "
                    "type '{}'".format(addon_types[opt_class_type])
                )

            identified_types.append(opt_class_type)

        new_class_name = (
            cls.__name__ +
            ''.join(class_.__name__ for class_ in addons)
        )
        mro_classes = tuple(list(addons) + [cls])
        new_class = type(new_class_name, mro_classes, {})
        new_class.main_class = cls

        return super(GradientDescent, new_class).__new__(new_class)

    def __init__(self, connection, options=None, floatX=None, **kwargs):
        if options is None:
            options = kwargs
        super(GradientDescent, self).__init__(connection, **options)

    def init_param_updates(self, layer, parameter):
        step = self.variables.step
        gradient = T.grad(self.variables.error_func, wrt=parameter)
        return [(parameter, parameter - step * gradient)]

    def class_name(self):
        return self.main_class.__name__

    def get_params(self, deep=False, with_connection=True):
        params = super(GradientDescent, self).get_params()
        if with_connection:
            params['connection'] = self.connection
        return params

    def __reduce__(self):
        parameters = self.get_params(with_connection=False)
        floatX = theano.config.floatX
        args = (self.connection, parameters, floatX)
        return (self.main_class, args)


class BatchSizeProperty(BoundedProperty):
    """
    Batch size property

    Parameters
    ----------
    {BoundedProperty.maxval}

    {BaseProperty.default}

    {BaseProperty.required}
    """
    expected_type = (type(None), int)
    fullbatch_identifiers = [None, -1, 'all', '*', 'full']

    def __init__(self, *args, **kwargs):
        super(BatchSizeProperty, self).__init__(minval=1, *args, **kwargs)

    def __set__(self, instance, value):
        if isinstance(value, six.string_types):
            value = value.lower()

        if value in self.fullbatch_identifiers:
            value = None

        super(BatchSizeProperty, self).__set__(instance, value)

    def validate(self, value):
        if value is not None:
            super(BatchSizeProperty, self).validate(value)


def iter_batches(n_samples, batch_size):
    """
    Iterates batch slices.

    Parameters
    ----------
    n_samples : int
        Number of samples. Number should be greater than ``0``.

    batch_size : int
        Mini-batch size. Number should be greater than ``0``.

    Yields
    ------
    object
        Batch slices.
    """
    n_batches = int(math.ceil(n_samples / batch_size))

    for batch_index in range(n_batches):
        yield slice(
            batch_index * batch_size,
            (batch_index + 1) * batch_size
        )


def cannot_divide_into_batches(data, batch_size):
    """
    Checkes whether data can be divided into at least
    two batches.

    Parameters
    ----------
    data : array-like
        Dataset.

    batch_size : int or None
        Size of the batch.

    Returns
    -------
    bool
    """
    if isinstance(data, (list, tuple)):
        # In case if network has more than one input
        data = data[0]

    n_samples = len(data)
    return batch_size is None or n_samples <= batch_size


def apply_batches(function, arguments, batch_size, description='',
                  show_progressbar=False, show_error_output=True):
    """
    Apply batches to a specified function.

    Parameters
    ----------
    function : func
        Function that accepts one or more positional arguments.
        Each of them should be an array-like variable that
        have exactly the same number of rows.

    arguments : tuple, list
        The arguemnts that will be provided to the function specified
        in the ``function`` argument.

    batch_size : int
        Mini-batch size.

    description : str
        Short description that will be displayed near the progressbar
        in verbose mode. Defaults to ``''`` (empty string).

    show_progressbar : bool
        ``True`` means that function will show progressbar in the
        terminal. Defaults to ``False``.

    show_error_output : bool
        Assumes that outputs from the function errors.
        ``True`` will show information in the progressbar.
        Error will be related to the last epoch.
        Defaults to ``True``.

    Returns
    -------
    list
        List of function outputs.
    """
    if not arguments:
        raise ValueError("The argument parameter should be list or "
                         "tuple with at least one element.")

    samples = arguments[0]
    n_samples = len(samples)
    batch_iterator = list(iter_batches(n_samples, batch_size))

    if show_progressbar:
        widgets = [
            progressbar.Timer(format='Time: %(elapsed)s'), ' |',
            progressbar.Percentage(),
            progressbar.Bar(),
            ' ', progressbar.ETA(),
        ]

        if show_error_output:
            widgets.extend([' | ', progressbar.DynamicMessage('error')])

        bar = progressbar.ProgressBar(
            widgets=widgets,
            max_value=len(batch_iterator),
            poll_interval=0.1,
        )
        bar.update(0)
    else:
        bar = progressbar.NullBar()

    outputs = []
    for i, batch in enumerate(batch_iterator):
        sliced_arguments = [argument[batch] for argument in arguments]

        output = function(*sliced_arguments)
        outputs.append(output)

        if show_error_output:
            bar.update(i, error=np.atleast_1d(output).item(0))
        else:
            bar.update(i)

    bar.fd.write('\r' + ' ' * bar.term_width + '\r')
    return outputs


def average_batch_errors(errors, n_samples, batch_size):
    """
    Computes average error per sample.

    Parameters
    ----------
    errors : list
        List of errors where each element is a average error
        per batch.

    n_samples : int
        Number of samples in the dataset.

    batch_size : int
        Mini-batch size.

    Returns
    -------
    float
        Average error per sample.
    """
    if batch_size is None:
        return errors[0]

    n_samples_in_final_batch = n_samples % batch_size

    if n_samples_in_final_batch == 0:
        return batch_size * sum(errors) / n_samples

    all_errors_without_last = errors[:-1]
    last_error = errors[-1]

    total_error = (
        sum(all_errors_without_last) * batch_size +
        last_error * n_samples_in_final_batch
    )
    average_error = total_error / n_samples
    return average_error


class MinibatchTrainingMixin(Configurable):
    """
    Mixin that helps to train network using mini-batches.

    Notes
    -----
    Works with ``BaseNetwork`` class.

    Parameters
    ----------
    batch_size : int or {{None, -1, 'all', '*', 'full'}}
        Set up min-batch size. If mini-batch size is equal
        to one of the values from the list (like ``full``) then
        it's just a batch that equal to number of samples.
        Defaults to ``128``.
    """
    batch_size = BatchSizeProperty(default=128)

    def apply_batches(self, function, input_data, arguments=(), description='',
                      show_progressbar=None, show_error_output=False):
        """
        Apply function per each mini-batch.

        Parameters
        ----------
        function : callable

        input_data : array-like
            First argument to the function that can be divided
            into mini-batches.

        arguments : tuple
            Additional arguments to the function.

        description : str
            Some description for the progressbar. Defaults to ``''``.

        show_progressbar : None or bool
            ``True``/``False`` will show/hide progressbar. If value
            is equal to ``None`` than progressbar will be visible in
            case if network expects to see logging after each
            training epoch.

        show_error_output : bool
            Assumes that outputs from the function errors.
            ``True`` will show information in the progressbar.
            Error will be related to the last epoch.

        Returns
        -------
        list
            List of outputs from the function. Each output is an
            object that ``function`` returned.
        """
        arguments = as_tuple(input_data, arguments)

        if cannot_divide_into_batches(input_data, self.batch_size):
            return [function(*arguments)]

        if show_progressbar is None:
            show_progressbar = (
                self.training and
                self.training.show_epoch == 1 and
                self.logs.enable
            )

        return apply_batches(
            function=function,
            arguments=arguments,
            batch_size=self.batch_size,

            description=description,
            show_progressbar=show_progressbar,
            show_error_output=show_error_output,
        )


def count_samples(input_data):
    """
    Count number of samples in the input data

    Parameters
    ----------
    input_data : array-like or list/tuple of array-like objects
        Input data to the network

    Returns
    -------
    int
        Number of samples in the input data.
    """
    if isinstance(input_data, (list, tuple)):
        return len(input_data[0])
    return len(input_data)


class MinibatchGradientDescent(GradientDescent, MinibatchTrainingMixin):
    """
    Mini-batch Gradient Descent algorithm.

    Parameters
    ----------
    {MinibatchTrainingMixin.Parameters}

    {GradientDescent.Parameters}

    Attributes
    ----------
    {GradientDescent.Attributes}

    Methods
    -------
    {GradientDescent.Methods}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> mgdnet = algorithms.MinibatchGradientDescent(
    ...     (2, 3, 1), batch_size=1
    ... )
    >>> mgdnet.train(x_train, y_train)

    See Also
    --------
    :network:`GradientDescent` : GradientDescent algorithm.
    """

    def train_epoch(self, input_train, target_train):
        """
        Train one epoch.

        Parameters
        ----------
        input_train : array-like
            Training input dataset.

        target_train : array-like
            Training target dataset.

        Returns
        -------
        float
            Training error.
        """
        errors = self.apply_batches(
            function=self.methods.train_epoch,
            input_data=input_train,
            arguments=as_tuple(target_train),

            description='Training batches',
            show_error_output=True,
        )
        return average_batch_errors(
            errors,
            n_samples=count_samples(input_train),
            batch_size=self.batch_size,
        )

    def prediction_error(self, input_data, target_data):
        """
        Check the prediction error for the specified input samples
        and their targets.

        Parameters
        ----------
        input_data : array-like
        target_data : array-like

        Returns
        -------
        float
            Prediction error.
        """
        input_data = self.format_input_data(input_data)
        target_data = self.format_target_data(target_data)

        errors = self.apply_batches(
            function=self.methods.prediction_error,
            input_data=input_data,
            arguments=as_tuple(target_data),

            description='Validation batches',
            show_error_output=True,
        )
        return average_batch_errors(
            errors,
            n_samples=count_samples(input_data),
            batch_size=self.batch_size,
        )

    def predict(self, input_data):
        """
        Makes a raw prediction.

        Parameters
        ----------
        input_data : array-like

        Returns
        -------
        array-like
        """
        outputs = self.apply_batches(
            function=self.methods.predict,
            input_data=self.format_input_data(input_data),

            description='Prediction batches',
            show_progressbar=True,
            show_error_output=False,
        )
        return np.concatenate(outputs, axis=0)



import numbers

import numpy as np
import theano.tensor as T

from neupy import init
from neupy.utils import number_type, as_tuple
from neupy.core.docs import SharedDocs


__all__ = ('BaseProperty', 'Property', 'ArrayProperty', 'BoundedProperty',
           'ProperFractionProperty', 'NumberProperty', 'IntProperty',
           'TypedListProperty', 'ChoiceProperty', 'WithdrawProperty',
           'ParameterProperty', 'CallableProperty')


class BaseProperty(SharedDocs):
    """
    Base class for properties.

    Parameters
    ----------
    default : object
        Default property value. Defaults to ``None``.

    required : bool
        If parameter equal to ``True`` and value is not defined
        after initialization then exception will be triggered.
        Defaults to ``False``.

    allow_none : bool
        When value is equal to ``True`` than ``None`` is a valid
        value for the parameter. Defaults to ``False``.

    Attributes
    ----------
    name : str or None
        Name of the property. ``None`` in case if name
        wasn't specified.

    expected_type : tuple or object
        Expected data types of the property.
    """
    expected_type = object

    def __init__(self, default=None, required=False, allow_none=False):
        self.name = None
        self.default = default
        self.required = required
        self.allow_none = allow_none

        if allow_none:
            self.expected_type = as_tuple(self.expected_type, type(None))

    def __set__(self, instance, value):
        if not self.allow_none or value is not None:
            self.validate(value)

        instance.__dict__[self.name] = value

    def __get__(self, instance, owner):
        if instance is None:
            return

        if self.default is not None and self.name not in instance.__dict__:
            self.__set__(instance, self.default)

        return instance.__dict__.get(self.name, None)

    def validate(self, value):
        """
        Validate properties value

        Parameters
        ----------
        value : object
        """
        if not isinstance(value, self.expected_type):
            availabe_types = self.expected_type

            if isinstance(availabe_types, (list, tuple)):
                availabe_types = ', '.join(t.__name__ for t in availabe_types)
            else:
                availabe_types = availabe_types.__name__

            raise TypeError("Invalid data type `{0}` for `{1}` property. "
                            "Expected types: {2}".format(
                                value.__class__.__name__, self.name,
                                availabe_types
                            ))

    def __repr__(self):
        classname = self.__class__.__name__

        if self.name is None:
            return '{}()'.format(classname)

        return '{}(name="{}")'.format(classname, self.name)


class WithdrawProperty(object):
    """
    Defines inherited property that needs to
    be withdrawed.

    Attributes
    ----------
    name : str or None
        Name of the property. ``None`` in case if name
        wasn't specified.
    """
    def __get__(self, instance, owner):
        # Remove itself, to make sure that instance doen't
        # have reference to this property. Instead user should
        # be able to see default value from the parent classes,
        # but not alowed to assign different value in __init__
        # method.
        #
        # Other part of functioality defined in the
        # ``ConfigMeta`` class.
        del self


class Property(BaseProperty):
    """
    Simple and flexible class that helps indetify properties with
    specified type.

    Parameters
    ----------
    expected_type : object
        Valid data types.

    {BaseProperty.Parameters}
    """
    def __init__(self, expected_type=object, *args, **kwargs):
        self.expected_type = expected_type
        super(Property, self).__init__(*args, **kwargs)


class ArrayProperty(BaseProperty):
    """
    Numpy array or matrix property.

    Parameters
    ----------
    {BaseProperty.Parameters}
    """
    expected_type = (np.ndarray, np.matrix)


class TypedListProperty(BaseProperty):
    """
    List property that contains specified element types.

    Parameters
    ----------
    n_elements : int
        Indentify fixed number of elements in list. ``None``
        value mean that list can contains any number of
        elements. Defaults to ``None``.

    element_type : object or tuple
        There are could be defined valid list elementy type
        or a bunch of them as tuple.

    {BaseProperty.Parameters}
    """
    expected_type = (list, tuple)

    def __init__(self, n_elements=None, element_type=int, *args, **kwargs):
        self.n_elements = n_elements
        self.element_type = element_type
        super(TypedListProperty, self).__init__(*args, **kwargs)

    def validate(self, value):
        super(TypedListProperty, self).validate(value)

        if self.n_elements is not None and len(value) != self.n_elements:
            raise ValueError("Expected list with {} variables"
                             "".format(self.n_elements))

        if not all(isinstance(v, self.element_type) for v in value):
            if isinstance(self.element_type, tuple):
                type_names = (type_.__name__ for type_ in self.element_type)
                element_type_name = ', '.join(type_names)
            else:
                element_type_name = self.element_type.__name__

            raise TypeError("Valid list element types for `{}` are: {}"
                            "".format(self.name, element_type_name))


class ChoiceProperty(BaseProperty):
    """
    Property that can have discrete number of properties.

    Parameters
    ----------
    choices : list, tuple or dict
        Stores all posible choices. Defines list of possible
        choices. If value specified as a dictionary than key
        would be just an alias to the expected value.

    {BaseProperty.Parameters}
    """
    choices = {}

    def __init__(self, choices, *args, **kwargs):
        super(ChoiceProperty, self).__init__(*args, **kwargs)
        self.choices = choices

        if isinstance(choices, (list, tuple, set)):
            self.choices = dict(zip(choices, choices))

        if not isinstance(self.choices, dict):
            class_name = self.choices.__class__.__name__
            raise ValueError("Choice properties can be only a `dict`, got "
                             "`{0}`".format(class_name))

        if not self.choices:
            raise ValueError("Must be at least one choice in property "
                             "`{0}`".format(self.name))

    def __set__(self, instance, value):
        if value in self.choices:
            return super(ChoiceProperty, self).__set__(instance, value)

        possible_choices = ", ".join(self.choices.keys())
        raise ValueError(
            "Wrong value `{0}` for property `{1}`. Available values: "
            "{2}".format(value, self.name, possible_choices)
        )

    def __get__(self, instance, owner):
        if instance is None:
            return

        choice_key = super(ChoiceProperty, self).__get__(instance, owner)
        return self.choices[choice_key]


class BoundedProperty(BaseProperty):
    """
    Number property that have specified numerical bounds.

    Parameters
    ----------
    minval : float
        Minimum possible value for the property.

    maxval : float
        Maximum possible value for the property.

    {BaseProperty.Parameters}
    """

    def __init__(self, minval=-np.inf, maxval=np.inf, *args, **kwargs):
        self.minval = minval
        self.maxval = maxval
        super(BoundedProperty, self).__init__(*args, **kwargs)

    def validate(self, value):
        super(BoundedProperty, self).validate(value)

        if not (self.minval <= value <= self.maxval):
            raise ValueError("Value `{}` should be between {} and {}"
                             "".format(self.name, self.minval, self.maxval))


class ProperFractionProperty(BoundedProperty):
    """
    Proper fraction property. Identify all possible numbers
    between zero and one.

    Parameters
    ----------
    {BaseProperty.Parameters}
    """
    expected_type = (float, int)

    def __init__(self, *args, **kwargs):
        super(ProperFractionProperty, self).__init__(minval=0, maxval=1,
                                                     *args, **kwargs)


class NumberProperty(BoundedProperty):
    """
    Float or integer number property.

    Parameters
    ----------
    {BoundedProperty.Parameters}
    """
    expected_type = number_type


class IntProperty(BoundedProperty):
    """
    Integer property.

    Parameters
    ----------
    {BoundedProperty.Parameters}
    """
    expected_type = (numbers.Integral, np.integer)

    def __set__(self, instance, value):
        if isinstance(value, float) and value.is_integer():
            value = int(value)
        super(IntProperty, self).__set__(instance, value)


class ParameterProperty(ArrayProperty):
    """
    In addition to Numpy arrays and matrix property support also
    Theano shared variables and NeuPy Initializers.

    Parameters
    ----------
    {ArrayProperty.Parameters}
    """
    expected_type = as_tuple(np.ndarray, T.sharedvar.SharedVariable,
                             T.Variable, init.Initializer, number_type)

    def __set__(self, instance, value):
        if isinstance(value, number_type):
            value = init.Constant(value)
        super(ParameterProperty, self).__set__(instance, value)


class CallableProperty(Property):
    """
    Property for callable objects.

    Parameters
    ----------
    {Property.Parameters}
    """
    def validate(self, value):
        if not callable(value):
            raise ValueError("The `{}` property expected to be "
                             "callable object.".format(self.name))
        super(CallableProperty, self).validate(value)



import theano.tensor as T
from theano.tensor.signal import pool

from neupy.utils import as_tuple
from neupy.core.properties import TypedListProperty, ChoiceProperty, Property
from neupy.exceptions import LayerConnectionError
from .base import BaseLayer
from .convolutions import StrideProperty


__all__ = ('MaxPooling', 'AveragePooling', 'Upscale', 'GlobalPooling')


class PaddingProperty(TypedListProperty):
    expected_type = as_tuple(TypedListProperty.expected_type, int)

    def __set__(self, instance, value):
        if isinstance(value, int):
            value = (value, value)
        super(PaddingProperty, self).__set__(instance, value)


def pooling_output_shape(dimension_size, pool_size, padding, stride,
                         ignore_border=True):
    """
    Computes output shape for pooling operation.

    Parameters
    ----------
    dimension_size : int
        Size of the dimension. Typically it's image's
        weight or height.

    filter_size : int
        Size of the pooling filter.

    padding : int
        Size of the zero-padding.

    stride : int
        Stride size.

    ignore_border : bool
        Defaults to ``True``.

    Returns
    -------
    int
    """
    if dimension_size is None:
        return None

    if ignore_border:
        output_size = dimension_size + 2 * padding - pool_size + 1
        output_size = (output_size + stride - 1) // stride

    elif stride >= pool_size:
        output_size = (dimension_size + stride - 1) // stride

    else:
        output_size = (dimension_size - pool_size + stride - 1) // stride
        output_size = max(1, output_size + 1)

    return output_size


class BasePooling(BaseLayer):
    """
    Base class for the pooling layers.

    Parameters
    ----------
    size : tuple with 2 integers
        Factor by which to downscale (vertical, horizontal).
        (2, 2) will halve the image in each dimension.

    stride : tuple or int.
        Stride size, which is the number of shifts over
        rows/cols to get the next pool region. If stride is
        None, it is considered equal to ds (no overlap on
        pooling regions).

    padding : tuple or int
        (pad_h, pad_w), pad zeros to extend beyond four borders of
        the images, pad_h is the size of the top and bottom margins,
        and pad_w is the size of the left and right margins.

    ignore_border : bool
        When ``True``, ``(5, 5)`` input with size ``(2, 2)``
        will generate a `(2, 2)` output. ``(3, 3)`` otherwise.
        Defaults to ``True``.

    {BaseLayer.Parameters}

    Methods
    -------
    {BaseLayer.Methods}

    Attributes
    ----------
    {BaseLayer.Attributes}
    """
    size = TypedListProperty(required=True, element_type=int)
    stride = StrideProperty(default=None)
    padding = PaddingProperty(default=0, element_type=int, n_elements=2)
    ignore_border = Property(default=True, expected_type=bool)

    def __init__(self, size, **options):
        super(BasePooling, self).__init__(size=size, **options)

        if not self.ignore_border and self.padding != (0, 0):
            raise ValueError("Cannot set padding parameter equal to {} while "
                             "``ignore_border`` is equal to ``False``"
                             "".format(self.padding))

    def validate(self, input_shape):
        if len(input_shape) != 3:
            raise LayerConnectionError(
                "Pooling layer expects an input with 3 "
                "dimensions, got {} with shape {}"
                "".format(len(input_shape), input_shape)
            )

    @property
    def output_shape(self):
        if self.input_shape is None:
            return None

        n_kernels, rows, cols = self.input_shape
        row_filter_size, col_filter_size = self.size

        stride = self.size if self.stride is None else self.stride

        row_stride, col_stride = stride
        row_padding, col_padding = self.padding

        output_rows = pooling_output_shape(rows, row_filter_size, row_padding,
                                           row_stride, self.ignore_border)
        output_cols = pooling_output_shape(cols, col_filter_size, col_padding,
                                           col_stride, self.ignore_border)

        return (n_kernels, output_rows, output_cols)

    def __repr__(self):
        return '{name}({size})'.format(name=self.__class__.__name__,
                                       size=self.size)


class MaxPooling(BasePooling):
    """
    Maximum pooling layer.

    Parameters
    ----------
    {BasePooling.Parameters}

    Methods
    -------
    {BasePooling.Methods}

    Attributes
    ----------
    {BasePooling.Attributes}

    Examples
    --------
    2D pooling

    >>> from neupy import layers
    >>>
    >>> network = layers.join(
    ...     layers.Input((3, 10, 10)),
    ...     layers.MaxPooling((2, 2)),
    ... )
    >>> network.output_shape
    (3, 5, 5)

    1D pooling

    >>> from neupy import layers
    >>>
    >>> network = layers.join(
    ...     layers.Input((10, 30)),
    ...     layers.Reshape((10, 30, 1)),
    ...     layers.MaxPooling((2, 1)),
    ... )
    >>> network.output_shape
    (10, 15, 1)
    """
    def output(self, input_value):
        return pool.pool_2d(input_value, ws=self.size, mode='max',
                            ignore_border=self.ignore_border,
                            stride=self.stride, pad=self.padding)


class AveragePooling(BasePooling):
    """
    Average pooling layer.

    Parameters
    ----------
    mode : {{``include_padding``, ``exclude_padding``}}
        Give a choice to include or exclude padding.
        Defaults to ``include_padding``.

    {BasePooling.Parameters}

    Methods
    -------
    {BasePooling.Methods}

    Attributes
    ----------
    {BasePooling.Attributes}

    Examples
    --------
    2D pooling

    >>> from neupy import layers
    >>>
    >>> network = layers.join(
    ...     layers.Input((3, 10, 10)),
    ...     layers.AveragePooling((2, 2)),
    ... )
    >>> network.output_shape
    (3, 5, 5)

    1D pooling

    >>> from neupy import layers
    >>>
    >>> network = layers.join(
    ...     layers.Input((10, 30)),
    ...     layers.Reshape((10, 30, 1)),
    ...     layers.AveragePooling((2, 1)),
    ... )
    >>> network.output_shape
    (10, 15, 1)
    """
    mode = ChoiceProperty(
        default='include_padding',
        choices={
            'include_padding': 'average_inc_pad',
            'exclude_padding': 'average_exc_pad'
        }
    )

    def output(self, input_value):
        return pool.pool_2d(input_value, ws=self.size, mode=self.mode,
                            ignore_border=self.ignore_border,
                            stride=self.stride, pad=self.padding)


class ScaleFactorProperty(TypedListProperty):
    """
    Defines sclaing factor for the Upscale layer.

    Parameters
    ----------
    {TypedListProperty.Parameters}
    """
    expected_type = (tuple, int)

    def __set__(self, instance, value):
        if isinstance(value, int):
            value = as_tuple(value, value)
        super(ScaleFactorProperty, self).__set__(instance, value)

    def validate(self, value):
        if any(element <= 0 for element in value):
            raise ValueError("Scale factor property accepts only positive "
                             "integer numbers.")
        super(ScaleFactorProperty, self).validate(value)


class Upscale(BaseLayer):
    """
    Upscales input over two axis (height and width).

    Parameters
    ----------
    scale : int or tuple with two int
        Scaling factor for the input value. In the tuple first
        parameter identifies scale of the height and the second
        one of the width.

    {BaseLayer.Parameters}

    Methods
    -------
    {BaseLayer.Methods}

    Attributes
    ----------
    {BaseLayer.Attributes}

    Examples
    --------
    >>> from neupy import layers
    >>>
    >>> network = layers.join(
    ...     layers.Input((3, 10, 10)),
    ...     layers.Upscale((2, 2)),
    ... )
    >>> network.output_shape
    (3, 20, 20)
    """
    scale = ScaleFactorProperty(required=True, n_elements=2)

    def __init__(self, scale, **options):
        super(Upscale, self).__init__(scale=scale, **options)

    def validate(self, input_shape):
        if len(input_shape) != 3:
            raise LayerConnectionError(
                "Upscale layer should have an input value with "
                "3 feature dimensions (channel, height, width)")

    @property
    def output_shape(self):
        if self.input_shape is None:
            return None

        channel, height, width = self.input_shape
        height_scale, width_scale = self.scale

        return (channel, height_scale * height, width_scale * width)

    def output(self, input_value):
        height_scale, width_scale = self.scale
        scaled_value = input_value

        if height_scale != 1:
            scaled_value = T.extra_ops.repeat(scaled_value, height_scale, 2)

        if width_scale != 1:
            scaled_value = T.extra_ops.repeat(scaled_value, width_scale, 3)

        return scaled_value


class GlobalPooling(BaseLayer):
    """
    Global pooling layer.

    Parameters
    ----------
    function : callable
        Function that aggregates over dimensions.
        Defaults to ``theano.tensor.mean``.

        .. code-block:: python

            def agg_func(x, axis=None):
                pass

    {BaseLayer.Parameters}

    Methods
    -------
    {BaseLayer.Methods}

    Attributes
    ----------
    {BaseLayer.Attributes}

    Examples
    --------
    >>> from neupy import layers
    >>>
    >>> network = layers.join(
    ...     layers.Input((16, 4, 4)),
    ...     layers.GlobalPooling(),
    ... )
    >>> network.output_shape
    (16,)
    """
    function = Property(default=T.mean)

    @property
    def output_shape(self):
        if self.input_shape is not None:
            return as_tuple(self.input_shape[0])

    def output(self, input_value):
        if input_value.ndim in (1, 2):
            return input_value

        agg_axis = range(2, input_value.ndim)
        return self.function(input_value, axis=list(agg_axis))



import numpy as np
import matplotlib.pyplot as plt

from neupy import algorithms, environment


environment.reproducible()
plt.style.use('ggplot')


input_data = np.array([
    [0.1961, 0.9806],
    [-0.1961, 0.9806],
    [0.9806, 0.1961],
    [0.9806, -0.1961],
    [-0.5812, -0.8137],
    [-0.8137, -0.5812],
])

sofmnet = algorithms.SOFM(
    n_inputs=2,
    n_outputs=3,

    step=0.5,
    show_epoch=20,
    shuffle_data=True,
    verbose=True,

    learning_radius=0,
    features_grid=(3, 1),
)

plt.plot(input_data.T[0:1, :], input_data.T[1:2, :], 'ko')
sofmnet.train(input_data, epochs=100)

print("> Start plotting")
plt.xlim(-1, 1.2)
plt.ylim(-1, 1.2)

plt.plot(sofmnet.weight[0:1, :], sofmnet.weight[1:2, :], 'bx')
plt.show()

for data in input_data:
    print(sofmnet.predict(np.reshape(data, (2, 1)).T))



from __future__ import division

import theano
import theano.tensor as T

from neupy.core.docs import shared_docs
from neupy.utils import asint


__all__ = ('mse', 'rmse', 'mae', 'msle', 'rmsle', 'binary_crossentropy',
           'categorical_crossentropy', 'binary_hinge', 'categorical_hinge')


def smallest_positive_number():
    """
    Function returns different nubmer for different
    ``theano.config.floatX`` values.

    * ``1e-7`` for 32-bit float
    * ``1e-16`` for 64-bit float

    Returns
    -------
    float
        Smallest positive float number.
    """
    float_type = theano.config.floatX
    epsilon_values = {
        'float16': 1e-3,
        'float32': 1e-7,
        'float64': 1e-16,
    }
    return epsilon_values[float_type]


def error_function(expected, predicted):
    """
    Parameters
    ----------
    expected : array-like, theano variable
    predicted : array-like, theano variable

    Returns
    -------
    array-like, theano variable
    """
    raise NotImplementedError


@shared_docs(error_function)
def mse(expected, predicted):
    """
    Mean squared error.

    .. math::
        mse(t, o) = mean((t - o) ^ 2)

    where :math:`t=expected` and :math:`o=predicted`

    Parameters
    ----------
    {error_function.expected}
    {error_function.predicted}

    Returns
    -------
    {error_function.Returns}
    """
    return T.square(predicted - expected).mean()


@shared_docs(error_function)
def rmse(expected, predicted):
    """
    Root mean squared error.

    .. math::
        rmse(t, o) = \\sqrt{{mean((t - o) ^ 2)}} = \\sqrt{{mse(t, 0)}}

    where :math:`t=expected` and :math:`o=predicted`

    Parameters
    ----------
    {error_function.expected}
    {error_function.predicted}

    Returns
    -------
    {error_function.Returns}
    """
    return T.sqrt(mse(expected, predicted))


@shared_docs(error_function)
def mae(expected, predicted):
    """
    Mean absolute error.

    .. math::
        mae(t, o) = mean(\\left| t - o \\right|)

    where :math:`t=expected` and :math:`o=predicted`

    Parameters
    ----------
    {error_function.expected}
    {error_function.predicted}

    Returns
    -------
    {error_function.Returns}
    """
    return T.abs_(expected - predicted).mean()


@shared_docs(error_function)
def msle(expected, predicted):
    """
    Mean squared logarithmic error.

    .. math::
        msle(t, o) = mean((\\log(t + 1) - \\log(o + 1)) ^ 2)

    where :math:`t=expected` and :math:`o=predicted`

    Parameters
    ----------
    {error_function.expected}
    {error_function.predicted}

    Returns
    -------
    {error_function.Returns}
    """
    squared_log = T.square(T.log(predicted + 1) - T.log(expected + 1))
    return squared_log.mean()


@shared_docs(error_function)
def rmsle(expected, predicted):
    """
    Root mean squared logarithmic error.

    .. math::
        rmsle(t, o) = \\sqrt{{
            mean((\\log(t + 1) - \\log(o + 1)) ^ 2)
        }} = \\sqrt{{msle(t, o)}}

    where :math:`t=expected` and :math:`o=predicted`

    Parameters
    ----------
    {error_function.expected}
    {error_function.predicted}

    Returns
    -------
    {error_function.Returns}
    """
    return T.sqrt(msle(expected, predicted))


@shared_docs(error_function)
def binary_crossentropy(expected, predicted):
    """
    Binary cross-entropy error.

    .. math::
        crossentropy(t, o) = -(t\\cdot log(o) + (1 - t) \\cdot log(1 - o))

    where :math:`t=expected` and :math:`o=predicted`

    Parameters
    ----------
    {error_function.expected}
    {error_function.predicted}

    Returns
    -------
    {error_function.Returns}
    """
    epsilon = smallest_positive_number()
    predicted = T.clip(predicted, epsilon, 1.0 - epsilon)
    return T.nnet.binary_crossentropy(predicted, expected).mean()


@shared_docs(error_function)
def categorical_crossentropy(expected, predicted):
    """
    Categorical cross-entropy error.

    Parameters
    ----------
    {error_function.expected}
    {error_function.predicted}

    Returns
    -------
    {error_function.Returns}
    """
    epsilon = smallest_positive_number()
    predicted = T.clip(predicted, epsilon, 1.0 - epsilon)
    return T.nnet.categorical_crossentropy(predicted, expected).mean()


def binary_hinge(expected, predicted, delta=1):
    """
    Computes the binary hinge loss between predictions
    and targets.

    .. math::
        hinge(t, o) = \\max(0, \\delta - t o)

    where :math:`t=expected` and :math:`o=predicted`

    Parameters
    ----------
    expected : Theano tensor
        Targets in {-1, 1} such as ground truth labels.

    predicted : Theano tensor
        Predictions in (-1, 1), such as hyprbolic tangent
        output of a neural network.

    delta : scalar
        The hinge loss margin. Defaults to ``1``.

    Returns
    -------
    Theano tensor
        An expression for the average binary hinge loss.

    Notes
    -----
    This is an alternative to the binary cross-entropy
    loss for binary classification problems.
    """
    error = T.nnet.relu(delta - predicted * expected)
    return error.mean()


def categorical_hinge(expected, predicted, delta=1):
    """
    Computes the multi-class hinge loss between
    predictions and targets.

    .. math::
        hinge_{i}(t, o) = \\max_{j \\not = o_i} (0, t_j - t_{o_i} + \\delta)

    Parameters
    ----------
    expected : Theano 2D tensor or 1D tensor
        Either a vector of int giving the correct class index
        per data point or a 2D tensor of one-hot encoding of
        the correct class in the same layout as predictions
        (non-binary targets in [0, 1] do not work!).

    predicted : Theano 2D tensor
        Predictions in (0, 1), such as softmax output of
        a neural network, with data points in rows and class
        probabilities in columns.

    delta : scalar
        The hinge loss margin. Defaults to ``1``.

    Returns
    -------
    Theano 1D tensor
        An expression for the average multi-class hinge loss.

    Notes
    -----
    This is an alternative to the categorical cross-entropy
    loss for multi-class classification problems.
    """
    n_classes = predicted.shape[1]

    if expected.ndim == (predicted.ndim - 1):
        expected = T.extra_ops.to_one_hot(asint(expected), n_classes)

    if expected.ndim != predicted.ndim:
        raise TypeError('Rank mismatch between expected and prediced values')

    invalid_class_indeces = expected.nonzero()
    valid_class_indeces = (1 - expected).nonzero()

    new_shape = (-1, n_classes - 1)
    rest = T.reshape(predicted[valid_class_indeces], new_shape)
    rest = T.max(rest, axis=1)

    corrects = predicted[invalid_class_indeces]
    error = T.nnet.relu(rest - corrects + delta)

    return error.mean()



from __future__ import division

from neupy.core.properties import IntProperty, NumberProperty
from .base import SingleStepConfigurable


__all__ = ('SearchThenConverge',)


class SearchThenConverge(SingleStepConfigurable):
    """
    Algorithm decrease learning step after each epoch.

    Parameters
    ----------
    reduction_freq : int
        The parameter controls the frequency reduction step
        with respect to epochs. Defaults to ``100`` epochs.
        Can't be less than ``1``. Less value mean that step
        decrease faster.

    rate_coefitient : float
        Second important parameter to control the rate of
        error reduction. Defaults to ``0.2``

    Warns
    -----
    {SingleStepConfigurable.Warns}

    Examples
    --------
    >>> from neupy import algorithms
    >>>
    >>> bpnet = algorithms.GradientDescent(
    ...     (2, 4, 1),
    ...     step=0.1,
    ...     verbose=False,
    ...     addons=[algorithms.SearchThenConverge]
    ... )
    >>>

    See Also
    --------
    :network:`StepDecay`
    """
    reduction_freq = IntProperty(minval=1, default=100)
    rate_coefitient = NumberProperty(default=0.2)

    def init_train_updates(self):
        updates = super(SearchThenConverge, self).init_train_updates()

        first_step = self.step
        reduction_freq = self.reduction_freq

        step = self.variables.step
        epoch = self.variables.epoch

        epoch_value = epoch / reduction_freq
        rated_value = 1 + (self.rate_coefitient / first_step) * epoch_value
        step_update_condition = (first_step * rated_value) / (
            rated_value + reduction_freq * epoch_value ** 2
        )

        updates.append((step, step_update_condition))
        return updates



import theano.tensor as T

from neupy import layers


__all__ = ('squeezenet',)


def Fire(s_1x1, e_1x1, e_3x3, name):
    return layers.join(
        layers.Convolution(
            (s_1x1, 1, 1),
            padding='half',
            name=name + '/squeeze1x1'
        ),
        layers.Relu(),
        [[
            layers.Convolution(
                (e_1x1, 1, 1),
                padding='half',
                name=name + '/expand1x1'
            ),
            layers.Relu(),
        ], [
            layers.Convolution(
                (e_3x3, 3, 3),
                padding='half',
                name=name + '/expand3x3'
            ),
            layers.Relu(),
        ]],
        layers.Concatenate(),
    )


def squeezenet():
    """
    SqueezeNet network architecture with random parameters.
    Parameters can be loaded using ``neupy.storage`` module.

    SqueezeNet has roughly 1.2 million parameters. It is almost
    50 times less than in AlexNet. Parameters can be stored as 5Mb
    file.

    Examples
    --------
    >>> from neupy import architectures
    >>> squeezenet = architectures.squeezenet()
    >>> squeezenet
    (3, 227, 227) -> [... 67 layers ...] -> 1000
    >>>
    >>> from neupy import algorithms
    >>> network = algorithms.Momentum(squeezenet)

    See Also
    --------
    :architecture:`vgg16` : VGG16 network
    :architecture:`vgg19` : VGG19 network
    :architecture:`alexnet` : AlexNet network
    :architecture:`resnet50` : ResNet50 network

    References
    ----------
    SqueezeNet: AlexNet-level accuracy with 50x fewer parameters
    and <0.5MB model size
    https://arxiv.org/abs/1602.07360
    """
    return layers.join(
        layers.Input((3, 227, 227)),

        layers.Convolution((96, 3, 3), stride=(2, 2),
                           padding='valid', name='conv1'),
        layers.Relu(),
        layers.MaxPooling((3, 3), stride=(2, 2)),

        Fire(16, 64, 64, name='fire2'),
        Fire(16, 64, 64, name='fire3'),
        Fire(32, 128, 128, name='fire4'),
        layers.MaxPooling((2, 2)),

        Fire(32, 128, 128, name='fire5'),
        Fire(48, 192, 192, name='fire6'),
        Fire(48, 192, 192, name='fire7'),
        Fire(64, 256, 256, name='fire8'),
        layers.MaxPooling((2, 2)),

        Fire(64, 256, 256, name='fire9'),
        layers.Dropout(0.5),

        layers.Convolution((1000, 1, 1), padding='valid', name='conv10'),
        layers.GlobalPooling(function=T.mean),
        layers.Reshape(),
        layers.Softmax(),
    )



import math

import numpy as np

from neupy.utils import format_data
from neupy.core.properties import Property
from .utils import bin2sign, hopfield_energy, step_function
from .base import DiscreteMemory


__all__ = ('DiscreteHopfieldNetwork',)


class DiscreteHopfieldNetwork(DiscreteMemory):
    """
    Discrete Hopfield Network. It can memorize binary samples
    and reconstruct them from corrupted samples.

    Notes
    -----
    - Works only with binary data. Input matrix should
      contain only zeros and ones.

    Parameters
    ----------
    {DiscreteMemory.mode}

    {DiscreteMemory.n_times}

    check_limit : bool
        Option enable a limit of patterns control for the
        network using logarithmically proportion rule.
        Defaults to ``True``.

        .. math::

            \\frac{{n_{{features}}}}{{2 \\cdot log_{{e}}(n_{{features}})}}

    Methods
    -------
    energy(input_data)
        Compute Discrete Hopfield Energy.

    train(input_data)
        Save input data pattern into the network memory.

    predict(input_data, n_times=None)
        Recover data from the memory using input pattern.
        For the prediction procedure you can control number
        of iterations. If you set up this value equal to ``None``
        then the value would be equal to the value that you
        set up for the property with the same name - ``n_times``.

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>>
    >>> def draw_bin_image(image_matrix):
    ...     for row in image_matrix.tolist():
    ...         print('| ' + ' '.join(' *'[val] for val in row))
    ...
    >>> zero = np.matrix([
    ...     0, 1, 1, 1, 0,
    ...     1, 0, 0, 0, 1,
    ...     1, 0, 0, 0, 1,
    ...     1, 0, 0, 0, 1,
    ...     1, 0, 0, 0, 1,
    ...     0, 1, 1, 1, 0
    ... ])
    >>>
    >>> one = np.matrix([
    ...     0, 1, 1, 0, 0,
    ...     0, 0, 1, 0, 0,
    ...     0, 0, 1, 0, 0,
    ...     0, 0, 1, 0, 0,
    ...     0, 0, 1, 0, 0,
    ...     0, 0, 1, 0, 0
    ... ])
    >>>
    >>> two = np.matrix([
    ...     1, 1, 1, 0, 0,
    ...     0, 0, 0, 1, 0,
    ...     0, 0, 0, 1, 0,
    ...     0, 1, 1, 0, 0,
    ...     1, 0, 0, 0, 0,
    ...     1, 1, 1, 1, 1,
    ... ])
    >>>
    >>> half_zero = np.matrix([
    ...     0, 1, 1, 1, 0,
    ...     1, 0, 0, 0, 1,
    ...     1, 0, 0, 0, 1,
    ...     0, 0, 0, 0, 0,
    ...     0, 0, 0, 0, 0,
    ...     0, 0, 0, 0, 0,
    ... ])
    >>>
    >>> draw_bin_image(zero.reshape((6, 5)))
    |   * * *
    | *       *
    | *       *
    | *       *
    | *       *
    |   * * *
    >>> draw_bin_image(half_zero.reshape((6, 5)))
    |   * * *
    | *       *
    | *       *
    |
    |
    |
    >>> data = np.concatenate([zero, one, two], axis=0)
    >>>
    >>> dhnet = algorithms.DiscreteHopfieldNetwork()
    >>> dhnet.train(data)
    >>>
    >>> result = dhnet.predict(half_zero)
    >>> draw_bin_image(result.reshape((6, 5)))
    |   * * *
    | *       *
    | *       *
    | *       *
    | *       *
    |   * * *

    See Also
    --------
    :ref:`password-recovery`: Password recovery with Discrete Hopfield Network.
    :ref:`discrete-hopfield-network`: Discrete Hopfield Network article.
    """
    check_limit = Property(default=True, expected_type=bool)

    def __init__(self, **options):
        super(DiscreteHopfieldNetwork, self).__init__(**options)
        self.n_memorized_samples = 0

    def train(self, input_data):
        self.discrete_validation(input_data)

        input_data = bin2sign(input_data)
        input_data = format_data(input_data, is_feature1d=False)

        n_rows, n_features = input_data.shape
        n_rows_after_update = self.n_memorized_samples + n_rows

        if self.check_limit:
            memory_limit = math.ceil(n_features / (2 * math.log(n_features)))

            if n_rows_after_update > memory_limit:
                raise ValueError("You can't memorize more than {0} "
                                 "samples".format(memory_limit))

        weight_shape = (n_features, n_features)

        if self.weight is None:
            self.weight = np.zeros(weight_shape, dtype=int)

        if self.weight.shape != weight_shape:
            n_features_expected = self.weight.shape[1]
            raise ValueError("Input data has invalid number of features. "
                             "Got {} features instead of {}."
                             "".format(n_features, n_features_expected))

        self.weight = input_data.T.dot(input_data)
        np.fill_diagonal(self.weight, np.zeros(len(self.weight)))
        self.n_memorized_samples = n_rows_after_update

    def predict(self, input_data, n_times=None):
        self.discrete_validation(input_data)
        input_data = format_data(bin2sign(input_data), is_feature1d=False)

        if self.mode == 'async':
            if n_times is None:
                n_times = self.n_times

            _, n_features = input_data.shape
            output_data = input_data

            for _ in range(n_times):
                position = np.random.randint(0, n_features - 1)
                raw_new_value = output_data.dot(self.weight[:, position])
                output_data[:, position] = np.sign(raw_new_value)
        else:
            output_data = input_data.dot(self.weight)

        return step_function(output_data).astype(int)

    def energy(self, input_data):
        self.discrete_validation(input_data)
        input_data = bin2sign(input_data)
        input_data = format_data(input_data, is_feature1d=False)
        n_rows, n_features = input_data.shape

        if n_rows == 1:
            return hopfield_energy(self.weight, input_data, input_data)

        output = np.zeros(n_rows)
        for i, row in enumerate(input_data):
            output[i] = hopfield_energy(self.weight, row, row)

        return output



import theano
import theano.tensor as T
import numpy as np

from neupy.utils import asfloat
from neupy.core.properties import ProperFractionProperty, Property
from .base import MinibatchGradientDescent


__all__ = ('Momentum',)


class Momentum(MinibatchGradientDescent):
    """
    Momentum algorithm.

    Parameters
    ----------
    momentum : float
        Control previous gradient ratio. Defaults to ``0.9``.

    nesterov : bool
        Instead of classic momentum computes Nesterov momentum.
        Defaults to ``False``.

    {MinibatchGradientDescent.Parameters}

    Attributes
    ----------
    {MinibatchGradientDescent.Attributes}

    Methods
    -------
    {MinibatchGradientDescent.Methods}

    Examples
    --------
    >>> import numpy as np
    >>> from neupy import algorithms
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> mnet = algorithms.Momentum((2, 3, 1))
    >>> mnet.train(x_train, y_train)

    See Also
    --------
    :network:`GradientDescent` : GradientDescent algorithm.
    """
    momentum = ProperFractionProperty(default=0.9)
    nesterov = Property(default=False, expected_type=bool)

    def init_param_updates(self, layer, parameter):
        step = self.variables.step

        parameter_shape = parameter.get_value().shape
        previous_velocity = theano.shared(
            name="{}/previous-velocity".format(parameter.name),
            value=asfloat(np.zeros(parameter_shape)),
        )

        gradient = T.grad(self.variables.error_func, wrt=parameter)
        velocity = self.momentum * previous_velocity - step * gradient

        if self.nesterov:
            velocity = self.momentum * velocity - step * gradient

        return [
            (parameter, parameter + velocity),
            (previous_velocity, velocity),
        ]
